# -*- coding: utf-8 -*-

### Copyright (C) 2005 Gustavo J. A. M. Carneiro
###
### This library is free software; you can redistribute it and/or
### modify it under the terms of the GNU Lesser General Public
### License as published by the Free Software Foundation; either
### version 2.1 of the License, or (at your option) any later version.
###
### This library is distributed in the hope that it will be useful,
### but WITHOUT ANY WARRANTY; without even the implied warranty of
### MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
### Lesser General Public License for more details.
###
### You should have received a copy of the GNU Lesser General Public
### License along with this library; if not, write to the Free Software
### Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

'''
A gtk.TextView-based renderer for XHTML-IM, as described in:
  http://www.jabber.org/jeps/jep-0071.html
'''
import gobject
import pango
import gtk
import xml.sax, xml.sax.handler
import re
import warnings
import urllib
from cStringIO import StringIO
import operator
import desktop

import stock
import dialog

__all__ = ['HtmlTextView']

whitespace_rx = re.compile("\\s+")

if gtk.gdk.screen_height_mm():
    ## pixels = points * display_resolution
    display_resolution = 0.3514598*(gtk.gdk.screen_height() /
                        float(gtk.gdk.screen_height_mm()))
else:
    # inaccurate, but better than letting the universe implode
    display_resolution = 1


def _parse_css_color(color):
    '''_parse_css_color(css_color) -> gtk.gdk.Color'''
    if color.startswith("rgb(") and color.endswith(')'):
        r, g, b = [int(c)*257 for c in color[4:-1].split(',')]
        return gtk.gdk.Color(r, g, b)
    else:
        return gtk.gdk.color_parse(color)
    

class HtmlHandler:
    
    def __init__(self, textview, startiter):
        self.textbuf = textview.get_buffer()
        self.textview = textview
        self.iter = startiter
        self.text = ''
        self.styles = [] # a gtk.TextTag or None, for each span level
        self.list_counters = [] # stack (top at head) of list
                                # counters, or None for unordered list

    def _parse_style_color(self, tag, value):
        color = _parse_css_color(value)
        tag.set_property("foreground-gdk", color)

    def _parse_style_background_color(self, tag, value):
        color = _parse_css_color(value)
        tag.set_property("background-gdk", color)
        if gtk.gtk_version >= (2, 8):
            tag.set_property("paragraph-background-gdk", color)


    if gtk.gtk_version >= (2, 8, 5) or gobject.pygtk_version >= (2, 8, 1):

        def _get_current_attributes(self):
            attrs = self.textview.get_default_attributes()
            self.iter.backward_char()
            self.iter.get_attributes(attrs)
            self.iter.forward_char()
            return attrs
        
    else:
        
        ## Workaround http://bugzilla.gnome.org/show_bug.cgi?id=317455
        def _get_current_style_attr(self, propname, comb_oper=None):
            tags = [tag for tag in self.styles if tag is not None]
            tags.reverse()
            is_set_name = propname + "-set"
            value = None
            for tag in tags:
                if tag.get_property(is_set_name):
                    if value is None:
                        value = tag.get_property(propname)
                        if comb_oper is None:
                            return value
                    else:
                        value = comb_oper(value, tag.get_property(propname))
            return value

        class _FakeAttrs(object):
            __slots__ = ("font", "font_scale")

        def _get_current_attributes(self):
            attrs = self._FakeAttrs()
            attrs.font_scale = self._get_current_style_attr("scale",
                                                            operator.mul)
            if attrs.font_scale is None:
                attrs.font_scale = 1.0
            attrs.font = self._get_current_style_attr("font-desc")
            if attrs.font is None:
                attrs.font = self.textview.style.font_desc
            return attrs


    def __parse_length_frac_size_allocate(self, textview, allocation,
                                          frac, callback, args):
        callback(allocation.width*frac, *args)

    def _parse_length(self, value, font_relative, callback, *args):
        '''Parse/calc length, converting to pixels, calls callback(length, *args)
        when the length is first computed or changes'''
        if value.endswith('%'):
            frac = float(value[:-1])/100
            if font_relative:
                attrs = self._get_current_attributes()
                font_size = attrs.font.get_size() / pango.SCALE
                callback(frac*display_resolution*font_size, *args)
            else:
                ## CSS says "Percentage values: refer to width of the closest
                ##           block-level ancestor"
                ## This is difficult/impossible to implement, so we use
                ## textview width instead; a reasonable approximation..
                alloc = self.textview.get_allocation()
                self.__parse_length_frac_size_allocate(self.textview, alloc,
                                                       frac, callback, args)
                self.textview.connect("size-allocate",
                                      self.__parse_length_frac_size_allocate,
                                      frac, callback, args)

        elif value.endswith('pt'): # points
            callback(float(value[:-2])*display_resolution, *args)

        elif value.endswith('em'): # ems, the height of the element's font
            attrs = self._get_current_attributes()
            font_size = attrs.font.get_size() / pango.SCALE
            callback(float(value[:-2])*display_resolution*font_size, *args)

        elif value.endswith('ex'): # x-height, ~ the height of the letter 'x'
            ## FIXME: figure out how to calculate this correctly
            ##        for now 'em' size is used as approximation
            attrs = self._get_current_attributes()
            font_size = attrs.font.get_size() / pango.SCALE
            callback(float(value[:-2])*display_resolution*font_size, *args)

        elif value.endswith('px'): # pixels
            callback(int(value[:-2]), *args)

        else:
            warnings.warn("Unable to parse length value '%s'" % value)
        
    def __parse_font_size_cb(length, tag):
        tag.set_property("size-points", length/display_resolution)
    __parse_font_size_cb = staticmethod(__parse_font_size_cb)

    def _parse_style_font_size(self, tag, value):
        try:
            scale = {
                "xx-small": pango.SCALE_XX_SMALL,
                "x-small": pango.SCALE_X_SMALL,
                "small": pango.SCALE_SMALL,
                "medium": pango.SCALE_MEDIUM,
                "large": pango.SCALE_LARGE,
                "x-large": pango.SCALE_X_LARGE,
                "xx-large": pango.SCALE_XX_LARGE,
                } [value]
        except KeyError:
            pass
        else:
            attrs = self._get_current_attributes()
            tag.set_property("scale", scale / attrs.font_scale)
            return
        if value == 'smaller':
            tag.set_property("scale", pango.SCALE_SMALL)
            return
        if value == 'larger':
            tag.set_property("scale", pango.SCALE_LARGE)
            return
        self._parse_length(value, True, self.__parse_font_size_cb, tag)

    def _parse_style_font_style(self, tag, value):
        try:
            style = {
                "normal": pango.STYLE_NORMAL,
                "italic": pango.STYLE_ITALIC,
                "oblique": pango.STYLE_OBLIQUE,
                } [value]
        except KeyError:
            warnings.warn("unknown font-style %s" % value)
        else:
            tag.set_property("style", style)

    def __frac_length_tag_cb(length, tag, propname):
        tag.set_property(propname, length)
    __frac_length_tag_cb = staticmethod(__frac_length_tag_cb)
        
    def _parse_style_margin_left(self, tag, value):
        self._parse_length(value, False, self.__frac_length_tag_cb,
                           tag, "left-margin")

    def _parse_style_margin_right(self, tag, value):
        self._parse_length(value, False, self.__frac_length_tag_cb,
                           tag, "right-margin")

    def _parse_style_font_weight(self, tag, value):
        ## TODO: missing 'bolder' and 'lighter'
        try:
            weight = {
                '100': pango.WEIGHT_ULTRALIGHT,
                '200': pango.WEIGHT_ULTRALIGHT,
                '300': pango.WEIGHT_LIGHT,
                '400': pango.WEIGHT_NORMAL,
                '500': pango.WEIGHT_NORMAL,
                '600': pango.WEIGHT_BOLD,
                '700': pango.WEIGHT_BOLD,
                '800': pango.WEIGHT_ULTRABOLD,
                '900': pango.WEIGHT_HEAVY,
                'normal': pango.WEIGHT_NORMAL,
                'bold': pango.WEIGHT_BOLD,
                } [value]
        except KeyError:
            warnings.warn("unknown font-style %s" % value)
        else:
            tag.set_property("weight", weight)

    def _parse_style_font_family(self, tag, value):
        tag.set_property("family", value)

    def _parse_style_text_align(self, tag, value):
        try:
            align = {
                'left': gtk.JUSTIFY_LEFT,
                'right': gtk.JUSTIFY_RIGHT,
                'center': gtk.JUSTIFY_CENTER,
                'justify': gtk.JUSTIFY_FILL,
                } [value]
        except KeyError:
            warnings.warn("Invalid text-align:%s requested" % value)
        else:
            tag.set_property("justification", align)
    
    def _parse_style_text_decoration(self, tag, value):
        tag.set_property("strikethrough", False)
        tag.set_property("underline", pango.UNDERLINE_NONE)
        
        for chunk in value.split(' '):
            if chunk == "underline":
                tag.set_property("underline", pango.UNDERLINE_SINGLE)
            elif chunk == "overline":
                warnings.warn("text-decoration:overline not implemented")
            elif chunk == "line-through":
                tag.set_property("strikethrough", True)
            elif value == "blink":
                warnings.warn("text-decoration:blink not implemented")
            else:
                warnings.warn("text-decoration:%s not implemented" % value)
        

    ## build a dictionary mapping styles to methods, for greater speed
    __style_methods = dict()
    for style in ["background-color", "color", "font-family", "font-size",
                  "font-style", "font-weight", "margin-left", "margin-right",
                  "text-align", "text-decoration"]:
        try:
            method = locals()["_parse_style_%s" % style.replace('-', '_')]
        except KeyError:
            warnings.warn("Style attribute '%s' not yet implemented" % style)
        else:
            __style_methods[style] = method
    del style
    ## --

    def _get_style_tags(self):
        return [tag for tag in self.styles if tag is not None]


    def _begin_span(self, style, tag=None):
        if style is None:
            self.styles.append(tag)
            return None
        if tag is None:
            tag = self.textbuf.create_tag()
        
        l = [item.split(':', 1) for item in style.split(';')]
        l = [ x for x in l if len( x ) == 2 ]
        for attr, val in l:
            attr = attr.strip().lower()
            # strip brake the font-family value
            # for example Sans Serif -> SansSerif
            val = val.lstrip().rstrip()
            try:
                method = self.__style_methods[attr]
            except KeyError:
                warnings.warn("Style attribute '%s' requested "
                              "but not yet implemented" % attr)
            else:
                method(self, tag, val)
        
        self.styles.append(tag)

    def _end_span(self):
        self.styles.pop(-1)

    def _insert_text(self, text):
        tags = self._get_style_tags()
        if tags:
            self.textbuf.insert_with_tags(self.iter, text, *tags)
        else:
            self.textbuf.insert(self.iter, text)
    
    def _flush_text(self):
        if not self.text: return
        self._insert_text(self.text.replace('\n', ''))
        self.text = ''

    def anchor_menu_copy_link(self, widget, href):
        clip = gtk.clipboard_get()
        clip.set_text(href)
        
    def anchor_menu_open_link(self, widget, href, type_):
        self.textview.emit("url-clicked", href, type_)

    def make_anchor_context_menu(self, event, href, type_):
        menu = gtk.Menu()
        
        menu_items = gtk.ImageMenuItem( _( "_Open link" ) )
        menu_items.set_image( gtk.image_new_from_stock( gtk.STOCK_NETWORK, gtk.ICON_SIZE_MENU ))
        menu.append(menu_items)
        menu_items.connect("activate", self.anchor_menu_open_link, href, type_)
        menu_items.show() 
                
        menu_items = gtk.ImageMenuItem( _( "_Copy link location" ))
        menu_items.set_image( gtk.image_new_from_stock( gtk.STOCK_COPY, gtk.ICON_SIZE_MENU ))
        menu.append(menu_items)
        menu_items.connect("activate", self.anchor_menu_copy_link, href)
        menu_items.show()      
            
        menu.popup(None, None, None, event.button, event.time)

    def _anchor_event(self, tag, textview, event, iter, href, type_):
        if event.type == gtk.gdk.BUTTON_PRESS and event.button == 1:
            self.textview.emit("url-clicked", href, type_)
            return True
        if event.type == gtk.gdk.BUTTON_PRESS and event.button == 3:
            self.make_anchor_context_menu( event, href, type_)
            return True
        return False
    
    def _object_event( self, widget, event, attrs ):
        callbacks = self.textview.customObjectsCallbacks
        if callbacks.has_key( attrs['type'] ) and callbacks[attrs['type']][1] != None:
            callbacks[attrs['type']][1]( widget, event, attrs )
        
    def char_data(self, content):
        self.text += content

    def start_element(self, name, attrs):
        self._flush_text()
        try:
            style = attrs['style']
        except KeyError:
            style = None

        tag = None
        if name == 'a':
            tag = self.textbuf.create_tag()
            tag.set_property('foreground', self.textview.linkColor)
            tag.set_property('underline', pango.UNDERLINE_SINGLE)
            try:
                type_ = attrs['type']
            except KeyError:
                type_ = None
            tag.connect('event', self._anchor_event, attrs['href'], type_)
            tag.is_anchor = True
        
        self._begin_span(style, tag)

        if name == 'br':
            pass # handled in endElement
        elif name == 'p':
            if not self.iter.starts_line():
                self._insert_text("\n")
        elif name == 'div':
            if not self.iter.starts_line():
                self._insert_text("\n")
        elif name == 'span':
            pass
        elif name == 'ul':
            if not self.iter.starts_line():
                self._insert_text("\n")
            self.list_counters.insert(0, None)
        elif name == 'ol':
            if not self.iter.starts_line():
                self._insert_text("\n")
            self.list_counters.insert(0, 0)
        elif name == 'li':
            if self.list_counters[0] is None:
                li_head = unichr(0x2022)
            else:
                self.list_counters[0] += 1
                li_head = "%i." % self.list_counters[0]
            self.text = ' '*len(self.list_counters)*4 + li_head + ' '
        elif name == 'img':
            try:
                ## Max image size = 10 MB (to try to prevent DoS)
                mem = urllib.urlopen(attrs['src']).read(10*1024*1024)
                ## Caveat: GdkPixbuf is known not to be safe to load
                ## images from network... this program is now potentially
                ## hackable ;)
                loader = gtk.gdk.PixbufLoader()
                loader.write(mem); loader.close()
                #pixbuf = loader.get_pixbuf()
                
                anchor = self.textbuf.create_child_anchor(self.iter)
                img = gtk.Image()
                img.set_from_animation(loader.get_animation())
                img.show()
                # TODO: add alt/tooltip with the special_text (a11y) 
                self.textview.add_child_at_anchor(img, anchor)
                
                
                
            except Exception, ex:
                pixbuf = None
                try:
                    alt = attrs['alt']
                except KeyError:
                    alt = "Broken image"
                    
        elif name == 'object':
            anchor = self.textbuf.create_child_anchor(self.iter)
            customObj = self.textview.setCustomObject( attrs['class'], type=attrs['type'] )
            
            event_box = gtk.EventBox()
            event_box.add( customObj )
            event_box.connect( 'event', self._object_event, attrs.copy() )
            event_box.set_visible_window( False )
            event_box.set_above_child( True )
            event_box.set_events(gtk.gdk.ALL_EVENTS_MASK)
            event_box.show_all()
            self.textview.add_child_at_anchor(event_box, anchor)

            
        elif name == 'body':
            pass
        elif name == 'a':
            pass
        else:
            warnings.warn("Unhandled element '%s'" % name)

    def end_element(self, name):
        self._flush_text()
        if name == 'p':
            if not self.iter.starts_line():
                self._insert_text("\n")
        elif name == 'div':
            if not self.iter.starts_line():
                self._insert_text("\n")
        elif name == 'span':
            pass
        elif name == 'br':
            self._insert_text("\n")
        elif name == 'ul':
            self.list_counters.pop()
        elif name == 'ol':
            self.list_counters.pop()
        elif name == 'li':
            self._insert_text("\n")
        elif name == 'img':
            pass
        elif name == 'object':
            pass
        elif name == 'body':
            pass
        elif name == 'a':
            pass
        else:
            warnings.warn("Unhandled element '%s'" % name)
        self._end_span()


class CustomEmoticonObject(object):
    def __init__(self, textview):
        self.imgs = []
        self.textview = textview
        self.pixbuf = None
        self.path = None
        
    def setNewImg(self, path=None):

        img = gtk.Image()
        img.set_from_stock(gtk.STOCK_MISSING_IMAGE, gtk.ICON_SIZE_SMALL_TOOLBAR)
        img.show()
        self.imgs.append( img )
        
        if path != None:
            self.setImgPath( path )

        self.updateImgs()
        return img
        
    def setImgPath( self, path ):
        self.path = path
        try:
            self.pixbuf = gtk.gdk.PixbufAnimation( path )
        except:
            pass
            
    def updateImgs( self ):
        if self.pixbuf is not None:
            self.textview.scrollLater()
            for img in self.imgs:
                img.set_from_animation(self.pixbuf)
        
    def getImgs( self ):
        return self.imgs


class HtmlTextView(gtk.TextView):
    __gtype_name__ = 'HtmlTextView'
    __gsignals__ = {
        'url-clicked': (gobject.SIGNAL_RUN_LAST, None, (str, str)), # href, type
    }
    
    def __init__(self, controller=None, buff=None, scrolledwin=None ):
        gtk.TextView.__init__(self, buff)
        self.set_wrap_mode(gtk.WRAP_WORD_CHAR)
        self.set_editable(False)
        self.set_pixels_above_lines(2)
        self.set_pixels_below_lines(2)
        
        self._last_mark = None
        self._changed_cursor = False
        self.linkColor = "#0000FF"

        self.connect("motion-notify-event", self.__motion_notify_event)
        self.connect("leave-notify-event", self.__leave_event)
        self.connect("enter-notify-event", self.__motion_notify_event)
        self.connect("url-clicked", self.on_link_clicked)
        if scrolledwin:
            scrolledwin.connect("expose-event", lambda x, y: self.queue_draw())

        self.controller = controller

        self.customObjects = {}
        self.customObjectsCallbacks = {
            'application/x-emesene-emoticon': (self.customEmoticon, 
                                                     self.customEmoticonEvent),
            'application/x-emesene-ink': (self.ink, None),
        }
        
    
    def customEmoticon( self, textview, id, path=None ):
        if self.customObjects.has_key( id ):
            return self.customObjects[ id ].setNewImg( path )
        else:
            customEmoticon = CustomEmoticonObject(textview)
            self.customObjects.update( {id:customEmoticon} )
            return customEmoticon.setNewImg( path )
            
    def customEmoticonEvent( self, img, event, attrs ):
        if not self.customObjects.has_key( attrs['class'] ):
            return
        path = self.customObjects[ attrs['class'] ].path
        if event.type == gtk.gdk.BUTTON_PRESS and event.button == 3:
            if path != None:
                menu = gtk.Menu()
                menu_items = gtk.ImageMenuItem( _( "_Save as ..." ) )
                menu_items.set_image( gtk.image_new_from_stock( gtk.STOCK_SAVE, gtk.ICON_SIZE_MENU ))
                menu.append(menu_items)
                menu_items.connect( 'activate', self.customEmoticonSave, path, attrs['data'] )
                menu_items.show()
                menu.popup(None, None, None, event.button, event.time)
                
    
    def customEmoticonSave( self, widget , path , shortcut ):
        def _on_ce_edit_cb(response, text=''):
            '''method called when the edition is done'''

            if response == stock.ACCEPT:
                if text:
                    success, msg = self.controller.customEmoticons.create(\
                        text, path, 1)
                
                    if not success:
                        dialog.error(msg)
                else:
                    dialog.error(_("Empty shortcut"))

        window = dialog.entry_window(_("New shortcut"), shortcut, 
            _on_ce_edit_cb, _("Shortcut"))
        window.show()
        
    def ink( self, textview, id, path=None ):
        img = gtk.Image()
        img.set_from_pixbuf( gtk.gdk.pixbuf_new_from_file( id ) )
        img.show()
        return img    
        
    def setCustomObject( self, id, path=None, type=None ):
        if self.customObjectsCallbacks.has_key( type ):
            return self.customObjectsCallbacks[type][0](self, id, path)
            
    def getCustomObjects( self ):
        return self.customObjects

    def on_link_clicked( self, view, url, type_ ):
        desktop.open(url)

    def __leave_event(self, widget, event):
        if self._changed_cursor:
            window = widget.get_window(gtk.TEXT_WINDOW_TEXT)
            window.set_cursor(gtk.gdk.Cursor(gtk.gdk.XTERM))
            self._changed_cursor = False
    
    def __motion_notify_event(self, widget, event):
        x, y, _ = widget.window.get_pointer()
        x, y = widget.window_to_buffer_coords(gtk.TEXT_WINDOW_TEXT, x, y)
        tags = widget.get_iter_at_location(x, y).get_tags()
        for tag in tags:
            if getattr(tag, 'is_anchor', False):
                is_over_anchor = True
                break
        else:
            is_over_anchor = False
        if not self._changed_cursor and is_over_anchor:
            window = widget.get_window(gtk.TEXT_WINDOW_TEXT)
            window.set_cursor(gtk.gdk.Cursor(gtk.gdk.HAND2))
            self._changed_cursor = True
        elif self._changed_cursor and not is_over_anchor:
            window = widget.get_window(gtk.TEXT_WINDOW_TEXT)
            window.set_cursor(gtk.gdk.Cursor(gtk.gdk.XTERM))
            self._changed_cursor = False
        return False
        
    def set_background(self, pixbuf):
        '''Sets a background pixbuf to this htmltextview
        If you are using this textview inside a ScrolledWindow, remember to
        pass the "scrolledwin" param to the constructor'''

        # converts the pixbuf to a server side pixmap
        pixmap = pixbuf.render_pixmap_and_mask()[0]

        # gets the text window
        textwnd = self.get_window(gtk.TEXT_WINDOW_TEXT)

        # sets background pixmap
        style = self.get_style()
        style.bg_pixmap[gtk.STATE_NORMAL] = pixmap
        style.set_background(textwnd, gtk.STATE_NORMAL)

        # this works too, but it doesn't apply background under scrollbars
        #textwnd.set_back_pixmap(pixmap, False)

    def display_html(self, html):
        buffer = self.get_buffer()
        eob = buffer.get_end_iter()
        parser = xml.parsers.expat.ParserCreate()
        handler = HtmlHandler(self, eob)
        parser.StartElementHandler = handler.start_element
        parser.EndElementHandler = handler.end_element
        parser.CharacterDataHandler = handler.char_data
        parser.Parse(html)

        if not eob.starts_line():
            buffer.insert(eob, "\n")

    def scrollToBottom(self, force=False):
        if not force and self.isScrollLocked():
            return False
        textbuffer = self.get_buffer()
        textiter = textbuffer.get_end_iter()
        mark = textbuffer.create_mark("end", textiter, False)
        self._last_mark = mark
        self.scroll_to_mark(mark, 0.05, True, 0.0, 1.0)
        textbuffer.place_cursor(textiter)
        return False

    def isScrollLocked(self):
        textiter = self.get_buffer().get_end_iter()
        if self._last_mark:
            rect = self.get_visible_rect()
            pos = rect.y + rect.height 
            if (pos + 25) < self.get_iter_location(textiter).y:
                return True
        return False
    
    def scrollLater(self):
        if not self.isScrollLocked():
            gobject.idle_add(self.scrollToBottom, True)

if __name__ == '__main__':
    sw = gtk.ScrolledWindow()
    htmlview = HtmlTextView(scrolledwin=sw)
    def url_cb(view, url, type_):
        print "url-clicked", url, type_
    htmlview.connect("url-clicked", url_cb)
        
    #~ htmlview.display_html('<div><span style="color: red; text-decoration:underline">Hello</span><br/>\n'
                          #~ '  <img src="http://images.slashdot.org/topics/topicsoftware.gif"/><br/>\n'
                          #~ '  <span style="font-size: 500%; font-family: serif">World</span>\n'
                          #~ '</div>\n')
    #~ htmlview.display_html("<br/>")
    #~ htmlview.display_html("""
      #~ <p style='font-size:large'>
        #~ <span style='font-style: italic'>O<span style='font-size:larger'>M</span>G</span>, 
        #~ I&apos;m <span style='color:green'>green</span>
        #~ with <span style='font-weight: bold'>envy</span>!
      #~ </p>
        #~ """)
    #~ htmlview.display_html("<br/>")
    #~ htmlview.display_html("""
    #~ <body xmlns='http://www.w3.org/1999/xhtml'>
      #~ <p>As Emerson said in his essay <span style='font-style: italic; background-color:cyan'>Self-Reliance</span>:</p>
      #~ <p style='margin-left: 5px; margin-right: 2%'>
        #~ &quot;A foolish consistency is the hobgoblin of little minds.&quot;
      #~ </p>
    #~ </body>
        #~ """)
    #~ htmlview.display_html("<br/>")
    #~ htmlview.display_html("""
    #~ <body xmlns='http://www.w3.org/1999/xhtml'>
      #~ <p style='text-align:center'>Hey, are you licensed to <a href='http://www.jabber.org/'>Jabber</a>?</p>
      #~ <p style='text-align:right'><img src='http://www.jabber.org/images/psa-license.jpg'
              #~ alt='A License to Jabber'
              #~ height='261'
              #~ width='537'/></p>
    #~ </body>
        #~ """)
    
    #~ htmlview.display_html("""
    #~ <body xmlns='http://www.w3.org/1999/xhtml'>
      #~ <ul style='background-color:rgb(120,140,100)'>
       #~ <li> One </li>
       #~ <li> Two </li>
       #~ <li> Three </li>
      #~ </ul>
    #~ </body>
        #~ """)
    #~ htmlview.display_html("""
    #~ <body xmlns='http://www.w3.org/1999/xhtml'>
      #~ <ol>
       #~ <li> One </li>
       #~ <li> Two </li>
       #~ <li> Three </li>
      #~ </ol>
    #~ </body>
        #~ """)
    htmlview.display_html('<body>If you are reading this, you may be ' \
        'interested in that useless text that came with htmltextview. ' \
        'Try uncommenting the code above this line.</body>')

    htmlview.show()

    sw.set_property("hscrollbar-policy", gtk.POLICY_AUTOMATIC)
    sw.set_property("vscrollbar-policy", gtk.POLICY_AUTOMATIC)
    sw.set_property("border-width", 0)
    sw.add(htmlview)
    sw.show()
    frame = gtk.Frame()
    frame.set_shadow_type(gtk.SHADOW_IN)
    frame.show()
    frame.add( sw )
    w = gtk.Window()
    w.add(frame)
    w.set_default_size(400, 300)
    w.show_all()

    # set background
    #pixbuf = gtk.gdk.pixbuf_new_from_file("Auto.jpg")
    #htmlview.set_background(pixbuf)

    w.connect("destroy", lambda w: gtk.main_quit())
    gtk.main()



# -*- coding: utf-8 -*-

#   This file is part of emesene.
#
#    Emesene is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    emesene is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with emesene; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
'''a module that define a class to contain all the information regarding a 
session
'''

class Session(Object.Object):
    '''a class that contain information and methods to handle a session'''

    def __init__(self, account, password, host, port):
        '''class constructor'''
        self.account = account
        self.password = password
        self.host = host
        self.port = port

        # set this on login
        self.contacts = None
        self.groups = None

        self.signal_add('login-successful', 0)
        # message
        self.signal_add('login-error', 1)
        # message
        self.signal_add('connection-error', 1)

    def login(self, account, password, stat):
        '''do login and set status'''
        raise NotImplementedError("This method isn't not implemented")

    def cancel_login(self):
        '''cancel the login process if posible on the protocol implementation
        '''
        raise NotImplementedError("This method isn't not implemented")

    def logout(self):
        '''do logout'''
        raise NotImplementedError("This method isn't not implemented")




# -*- coding: utf-8 -*-

#   This file is part of emesene.
#
#    Emesene is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    emesene is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with emesene; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

VERSION = '0.3'

import os

import CurrentSong

error = False
dcopext = None
pydcop = None
try:
    import dcopext
except:
    try:
        import pydcop
    except:
        error = True

class Amarok( CurrentSong.CurrentSong ):

    def __init__( self ):
        CurrentSong.CurrentSong.__init__( self )

        self.amarok = None
        if dcopext:
            self.client = dcopext.DCOPClient()
            self.client.attach()
        self.check()
        
    def _isServerRunning( self ):
        try:
            assert os.system( "ps aux|grep dcopserver|grep -v grep" ) == 0
        except:
            return False
        else:
            return True
        
    def isRunning( self ):
        if pydcop:
            return bool(self._isServerRunning() and "amarok" in pydcop.apps())
        elif dcopext:
            return bool(self.client.isApplicationRegistered('amarok'))

    def isPlaying( self ):
        if self.amarok:
            if pydcop and self.amarok.player.isPlaying():
                return True
            elif dcopext and self.amarok.player.isPlaying()[1]:
                return True
        
        return False

    def getStatus( self ):
        '''check if everything is OK to start the plugin
        return a tuple whith a boolean and a message
        if OK -> ( True , 'some message' )
        else -> ( False , 'error message' )'''
        global error

        if error:
            return ( False, _( "Can't import dcopext or python-dcop!" ) )

        if pydcop and not self._isServerRunning():
            return ( False, _( "dcop not running" ) )

        if os.name != 'posix':
            return ( False, _( 'This plugin only works in posix systems' ) )

        if pydcop and not self.is_on_path( 'dcop' ):
            return ( False, _( 'Dcop not found' ) )

        return ( True, 'Ok' )

    def check( self ):
        global error

        if error:
            return False

        if self.amarok == None and self.isRunning():
            if dcopext:
                self.amarok = dcopext.DCOPApp("amarok", self.client)
            elif pydcop:
                self.amarok = pydcop.DCOPApplication( "amarok" )

        if self.isPlaying():
            if dcopext:
                artist = self.amarok.player.artist()[1]
                title = self.amarok.player.title()[1]
                album = self.amarok.player.album()[1]
            elif pydcop:
                artist = self.amarok.player.artist()
                title = self.amarok.player.title()
                album = self.amarok.player.album()
                
            if self.artist != artist or self.title != title:
                self.artist = artist
                self.title = title
                self.album = album

                return True

        return False



# -*- coding: utf-8 -*-

#   This file is part of emesene.
#
#    Emesene is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    emesene is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with emesene; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import gtk
import Queue
import thread
import gobject

class Plugin( object ):
    '''Base class to make a plugin'''
    
    def __init__( self, controller, msn, period = None ):
        '''Contructor, the period parameter is the time between calls to 
        the function, in milliseconds, if period = None then the plugin 
        is asynchronous.'''
        
        self.controller = controller
        self.msn = msn
        
        self.period = period
        
        self.name = ''
        self.description = ''
        # name : mail
        self.authors = {}
        self.website = ''
        self.enabled = False
        
    def check( self ):
        '''check if everything is OK to start the plugin
        return a tuple whith a boolean and a message
        if OK -> ( True , 'some message' )
        else -> ( False , 'error message' )'''
        
        pass
    
    def start( self ):
        '''start the plugin'''
        
        pass
    
    def stop( self ):
        '''stop the plugin'''
        
        pass
    
    def action( self ):
        '''the method that is called every self.period milliseconds'''
        
        pass
    
    def connect( self, signal, callback ):
        '''connect a signal to a callback'''
        
        return self.msn.connect( signal, callback )
        
    def disconnect( self, callbackId ):
        '''disconnect a signal to a callback'''
        
        self.msn.disconnect( callbackId )
        
    def isEnabled( self ):
        '''return True if the plugin is running'''
        
        return self.enabled
    
    # TODO: iter?
    def getOpenConversations( self ):
        if self.controller.conversationManager == None:
            return ()
        conversations = self.controller.conversationManager.conversations
        return [conversation for (window, conversation) in conversations]

class Function( object ):
    '''a class that represent a function that will run
    on a thread and call a callback with the return value'''

    def __init__( self, function, callback ):
        self.function = function
        self.callback = callback
        self.queue = Queue.Queue( 0 )

    def __call__( self, *args ):
        '''when the object is called we start a thread
        with the function and parameters and start the
        check function that check every 0.2 seconds if
        the thread has ended (there is something in the
        queue), then call the callback with the result'''

        thread.start_new_thread( self._function, args )
        gobject.timeout_add( 200 , self.check )

    def _function( self, *args ):
        '''our function call the parameter function
        with the args and put the return value in the
        queue'''

        self.queue.put( self.function( *args ) )

    def check( self ):
        '''check there is something in the queue 
        that means that self.function has ended
        then call the callback with the result'''

        try:
            result = self.queue.get( True, 0.01 )
            self.callback( result )
            return False
        except Queue.Empty:
            return True
    
class Option( object ):
    '''a type of option to use by the plugins'''
    
    def __init__( self, name, optionType, label, description, value, \
                  options=None, changedcb=None ):
        '''constructor
        name: represent the name of the variable
        optionType: the type expected (str, int, float, bool, list,passwd)
        label: the text that will appear in the label
        descrption: the text in the tooltip
        value: the default value when show and the input value when returned'''
        
        self.name = name
        self.optionType = optionType
        self.label = label
        self.description = description
        self.value = value
        self.options = options
        self.changedcb = changedcb

        self.widget = None

class ConfigList(gtk.TreeView):
    '''A treeview that displays a dict and allows changing the values'''

    def __init__(self, data):
        self.model = gtk.ListStore(str, str)
        gtk.TreeView.__init__(self, self.model)
        self.refresh(data)
        
        crtkey = gtk.CellRendererText()
        crtvalue = gtk.CellRendererText()
        crtvalue.set_property('editable', True)
        crtvalue.connect('edited', self.changed, self.model)
        
        col1 = gtk.TreeViewColumn(_("Key"), crtkey, text=0)
        col2 = gtk.TreeViewColumn(_("Value"), crtvalue, text=1)
        self.append_column(col1)
        self.append_column(col2)

    def get_data( self ):
        return self.data
    
    def refresh( self, data ):
        '''Adds dict data to the model'''
        self.data = data
        self.model.clear()
        for item in self.data.keys():
            self.model.append((item, self.data[item]))

    def changed( self, cell, path, new_text, model ):
        '''a value is changed to new_text
        model[path] is the row
        model[path][0] is the key
        model[path][1] is the stored value
        self.data contains the _return_ values'''
        
        model[path][1] = new_text
        key = model[path][0]
        if key in self.data:
            self.data[key] = new_text
        else:
            # key not in data..?
            pass

class ConfigWindow( gtk.Dialog ):
    '''build a generic window with configuration parameters to fill'''
    
    def __init__( self, title, optionList ):
        '''constructor'''
        
        gtk.Dialog.__init__( self, title, None,
            gtk.DIALOG_MODAL | gtk.DIALOG_DESTROY_WITH_PARENT,
            (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL, \
             gtk.STOCK_OK, gtk.RESPONSE_OK) )
        self.set_position(gtk.WIN_POS_MOUSE)
        self.set_border_width( 6 )
        
        vbox = self.vbox
        vbox.set_spacing( 12 )
        
        self.options = {}
        self.widgets = {}
        for i in optionList:
            self.options[ i.name ] = i
            hbox = gtk.HBox()
            if i.optionType == bool:
                check = gtk.CheckButton( i.label )

                if type( i.value ) == bool:
                    check.set_active( i.value )
                
                self.widgets[ i.name ] = check.get_active
                if i.changedcb:
                    check.connect( 'toggled', i.changedcb )
                hbox.pack_start( check )

            elif i.optionType in ('list', list) and type(i.options) == list:
                label = gtk.Label( i.label )
                label.set_justify(gtk.JUSTIFY_LEFT)
                label.set_width_chars( 20 )
                    
                combo = gtk.ComboBox()
                
                liststore = gtk.ListStore(str)
                
                cellRendererText = gtk.CellRendererText()
                combo.pack_start(cellRendererText)
                combo.add_attribute(cellRendererText, 'text', 0)
                combo.set_model(liststore)
                
                self.widgets[ i.name ] = combo.get_active_text
                hbox.pack_start( label )
                hbox.pack_start( combo )
                
                count = 0
                done = False
                for j in i.options:
                    if not done:
                        if j == i.value:
                            done = True
                        else:
                            count += 1
                        
                    liststore.append( [ str( j ) ] )
                
                if not done:
                    combo.set_active( 0 )
                else:   
                    combo.set_active( count )
                if i.changedcb:
                    combo.connect( 'changed', i.changedcb )

            elif i.optionType == dict:
                label = gtk.Label( i.label )
                label.set_justify(gtk.JUSTIFY_LEFT)
                label.set_width_chars( 20 )

                cfglist = ConfigList(i.options)
                cfglist.set_size_request(-1, 200)

                i.widget = cfglist
                self.widgets[ i.name ] = cfglist.get_data
                
                scrolledWindow = gtk.ScrolledWindow()
                scrolledWindow.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
                scrolledWindow.add(cfglist)

                vbox.pack_start(label)
                vbox.pack_start(scrolledWindow)
            elif i.optionType == 'passwd':
                label = gtk.Label( i.label )
                label.set_justify(gtk.JUSTIFY_LEFT)
                label.set_width_chars( 20 )
                
                entry = gtk.Entry()
                entry.set_visibility(False)
                entry.set_text( str( i.value) )
                
                self.widgets[ i.name ] = entry.get_text
                
                hbox.pack_start( label )
                hbox.pack_start( entry )

            elif i.optionType == gtk.Widget:
                if i.label:
                    label = gtk.Label( i.label )
                    label.set_justify(gtk.JUSTIFY_LEFT)
                    label.set_width_chars( 20 )
                    hbox.pack_start( label )
                hbox.pack_start( i.value )
                self.widgets[ i.name ] = lambda: None
                
            else:
                label = gtk.Label( i.label )
                label.set_justify(gtk.JUSTIFY_LEFT)
                label.set_width_chars( 20 )
                
                entry = gtk.Entry()
                entry.set_text( str( i.value ) )
            
                if i.changedcb:
                    entry.connect( 'changed', i.changedcb )

                self.widgets[ i.name ] = entry.get_text
                hbox.pack_start( label )
                hbox.pack_start( entry )
                
            vbox.pack_start( hbox )

        vbox.show_all()
        self.connect( 'delete-event', self.close )
        
    def run( self, *args ):
        response = gtk.Dialog.run( self )
        self.hide()

        if response == gtk.RESPONSE_OK:
            for i in self.options.keys():
                self.options[ i ].value = self.widgets[ i ]()
            return self.options
        else:
            return None

    def close( self, *args ):
        self.response( gtk.RESPONSE_CLOSE )

if __name__ == '__main__':
    import time
    import Plugin

    def funcion(a,b):
        time.sleep(5)
        return a+b


    def callback( suma ):
        print suma

    f = Plugin.Function( funcion, callback )
    f( 2,3 )
    import gtk
    gtk.main()

    



# -*- coding: utf-8 -*-
#
#   This file is part of emesene.
#
#    Emesene is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    Emesene is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with emesene; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#

'''Converts a gtk.gdk.Pixbuf into a GifEncoder instance'''

import pygif

import StringIO
import gtk

def convert(pixbuf, enc=None):
    '''Parses gtk.gdk.Pixbuf pixels and returns a GifEncoder instance'''

    if enc is None:
        enc = pygif.GifEncoder()

    enc.ls_width = pixbuf.get_width()
    enc.ls_height = pixbuf.get_height()
    bpp = pixbuf.get_n_channels()
    assert pixbuf.get_bits_per_sample() == 8
    
    # we have two palletes
    # the "definitive", tuple based pallete
    # and a quick search string pallete
    enc.pallete = []
    string_pallete = []

    raw_pixels = StringIO.StringIO(pixbuf.get_pixels())
    out_pixels = []

    while True:
        rgb = raw_pixels.read(bpp)
        if len(rgb) != bpp:
            break
        
        red, green, blue = ord(rgb[0]), ord(rgb[1]), ord(rgb[2])
        # uncomment the following to have ASCII art debug :P
        #sys.stdout.write(hex(red)[2])
        
        if rgb not in string_pallete:
            index = len(enc.pallete)
            string_pallete.append(rgb)
            enc.pallete.append((red, green, blue))
        else:
            index = string_pallete.index(rgb)

        out_pixels.append(index)
    del string_pallete

    enc.global_color_table_size = len(enc.pallete)
    enc.color_table_flag = True
    enc.color_resolution = 7 # 256 colors
    enc.build_flags()
    
    image = enc.new_image()
    image.codesize, image.lzwcode = enc.lzw_encode(out_pixels)
    return enc

def main():
    '''runs a simple test with vampire.gif'''
    pixbuf = gtk.gdk.pixbuf_new_from_file("vampire.gif")
    enc = convert(pixbuf)
    enc.write("vampire_pixbuf.gif")

if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

#   This file is part of emesene.
#
#    Emesene is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    emesene is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with emesene; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import gtk
import desktop
import StatusMenu
import PreferenceWindow

import stock
import dialog

class MainMenu(gtk.MenuBar):
    '''this class represent the main menu in the main window'''

    def __init__(self, controller, window_type, accelGroup):
        '''constructor'''
        gtk.MenuBar.__init__(self)

        self.controller = controller
        self.config = self.controller.config
        self.theme = self.controller.theme

        fileMenu = gtk.Menu()
        fileMenuItem = self.newImageMenuItem(_('_File'))
        fileMenuItem.set_submenu(fileMenu)

        # If we're in login window we don't need the View 
        # or the Actions Menu
        if window_type == 'userlist':
            viewMenu = gtk.Menu()
            viewMenuItem = self.newImageMenuItem(_('_View'))
            viewMenuItem.set_submenu(viewMenu)
            
            actionsMenu = gtk.Menu()
            actionsMenuItem = self.newImageMenuItem(_('_Actions'))
            actionsMenuItem.connect('activate', self.on_actions_activate)
            actionsMenuItem.set_submenu(actionsMenu)

        optionsMenu = gtk.Menu()
        optionsMenuItem = self.newImageMenuItem(_('_Options'))
        optionsMenuItem.set_submenu(optionsMenu)

        helpMenu = gtk.Menu()
        helpMenuItem = self.newImageMenuItem(_('_Help'))
        helpMenuItem.set_submenu(helpMenu)

        # ---------------------- FILE MENU --------------------------------

        # In login window only quit option is usefull
        if window_type == 'userlist':
            statusMenuItem = self.newImageMenuItem(_('_Status'))
            statusMenu = StatusMenu.StatusMenu(self.controller)

            statusMenuItem.set_submenu(statusMenu)
            fileMenu.add(statusMenuItem)
            
            fileMenu.add(gtk.SeparatorMenuItem())
            
            disconnect = self.newStockImageMenuItem(gtk.STOCK_DISCONNECT)
            fileMenu.add(disconnect)
        else:
            getliveMenuItem = self.newImageMenuItem(_('_New account'), \
                                                    gtk.STOCK_NEW)
            fileMenu.add(getliveMenuItem)
            fileMenu.add(gtk.SeparatorMenuItem())

        # Quit option is always visible
        quitMenuItem = self.newStockImageMenuItem(gtk.STOCK_QUIT)
        quitMenuItem.add_accelerator('activate', accelGroup, ord('Q'), \
                                 gtk.gdk.CONTROL_MASK, gtk.ACCEL_VISIBLE)
        fileMenu.add(quitMenuItem)
        
        accelGroup.connect_group(ord('Q'), gtk.gdk.CONTROL_MASK, \
            gtk.ACCEL_LOCKED, self.on_quit_activate)

        # -------------------- VIEW MENU ----------------------------------
        
        # Action Menu is useless in login window
        if window_type == 'userlist':
            self.orderByGroup = gtk.RadioMenuItem(None, \
                                                      _('Order by _group'))
            self.orderByStatus = gtk.RadioMenuItem(self.orderByGroup, \
                                                      _('Order by _status'))
            if not self.config.user['orderByStatus']:
                self.orderByGroup.set_active(True)
            else:
                self.orderByStatus.set_active(True)
                
            viewMenu.add(self.orderByGroup)
            viewMenu.add(self.orderByStatus)
            
            viewMenu.add(gtk.SeparatorMenuItem())

            self.showByNick = self.newCheckMenuItem(_('Show by _nick'), \
                    self.config.user['showByNick'])
            self.showOffline = self.newCheckMenuItem(_('Show _offline'), \
                    self.config.user['showOffline'])
            self.showEmptyGroups = self.newCheckMenuItem(
             _('Show _empty groups'), self.config.user['showEmptyGroups'])
            self.showCountContact = self.newCheckMenuItem(
                _('Show _contact count'), 
                self.config.user['showCountContact'])

            viewMenu.add(self.showByNick)
            viewMenu.add(self.showOffline)
            viewMenu.add(self.showEmptyGroups)
            viewMenu.add(self.showCountContact)

        # ------------------- ACTIONS MENU --------------------------------

        # Action Menu is useless in login window
        if window_type == 'userlist':
            # Add user, group
            addUserMenuItem = self.newImageMenuItem(_('_Add contact...'), \
                                                     gtk.STOCK_ADD)
            actionsMenu.add(addUserMenuItem)

            addGroupMenuItem = self.newImageMenuItem(_('Add _group...'),
                gtk.STOCK_ADD)
            actionsMenu.add(addGroupMenuItem)
            
            actionsMenu.add(gtk.SeparatorMenuItem())
            
            # Contacts action group: remove, block, unblock
            contactsActionGroup = gtk.ActionGroup('contacts')
            self.contactsActionGroup = contactsActionGroup
            
            setAliasAction = gtk.Action('setAlias', 
                _('_Set contact alias...'), None, gtk.STOCK_EDIT)
            blockAction = gtk.Action('block', _('_Block'), \
                                        None, gtk.STOCK_STOP)
            unblockAction = gtk.Action('unblock', _('_Unblock'), \
                                        None, gtk.STOCK_APPLY)
            moveAction = gtk.Action('move', _('M_ove to group'), \
                                     None, gtk.STOCK_REDO)
            deleteAction = gtk.Action('delete', _('_Remove contact'), \
                                        None, gtk.STOCK_DELETE)
            
            self.blockAction = blockAction
            self.unblockAction = unblockAction
            
            contactsActionGroup.add_action(setAliasAction)
            actionsMenu.add(setAliasAction.create_menu_item())
            contactsActionGroup.add_action(blockAction)
            actionsMenu.add(blockAction.create_menu_item())
            contactsActionGroup.add_action(unblockAction)
            actionsMenu.add(unblockAction.create_menu_item())
            contactsActionGroup.add_action(moveAction)
            moveMenuItem = moveAction.create_menu_item()
            actionsMenu.add(moveMenuItem)
            contactsActionGroup.add_action(deleteAction)
            actionsMenu.add(deleteAction.create_menu_item())
            
            # ------------ move to submenu ----------
            moveMenu = gtk.Menu()

            for i in self.controller.msn.getGroupNames():
                menuItem = gtk.MenuItem (i)
                moveMenu.add(menuItem)
                menuItem.connect('activate', self.on_move_to_activate, i)
    
            moveMenuItem.set_submenu(moveMenu)
            moveMenuItem.show_all()
            moveMenu.show_all()
            # ---------------------------------------

            actionsMenu.add(gtk.SeparatorMenuItem())
            
            # Groups action group: remove, rename
            groupsActionGroup = gtk.ActionGroup('groups')
            self.groupsActionGroup = groupsActionGroup
            
            deleteGAction = gtk.Action('remove', _('Re_move group'), \
                                        None, gtk.STOCK_DELETE)
            renameAction = gtk.Action('rename', _('Re_name group...'), \
                                        None, gtk.STOCK_EDIT)
            
            groupsActionGroup.add_action(renameAction)
            actionsMenu.add(renameAction.create_menu_item())
            groupsActionGroup.add_action(deleteGAction)
            actionsMenu.add(deleteGAction.create_menu_item())

        # ------------------------ OPTIONS MENU ---------------------------

        if window_type == 'userlist':            
            changeNickMenuItem = self.newImageMenuItem(\
                _('_Change nick...'), gtk.STOCK_EDIT)
            changeAvatarMenuItem = self.newImageMenuItem(\
                _('Change _display picture...'), gtk.STOCK_EDIT)
            optionsMenu.add(changeNickMenuItem)
            optionsMenu.add(changeAvatarMenuItem)
            
            optionsMenu.add(gtk.SeparatorMenuItem())
            
            setAutoReplyMenuItem = self.newImageMenuItem(\
                _('Edit a_utoreply...'), gtk.STOCK_EDIT)

            self.activateAutoReply = self.newCheckMenuItem(\
                _('Activate au_toreply'), self.config.user['autoReply'])

            optionsMenu.add(self.activateAutoReply)
            optionsMenu.add(setAutoReplyMenuItem)
            
            optionsMenu.add(gtk.SeparatorMenuItem())
            
            pluginMenuItem = self.newImageMenuItem(_('P_lugins'), \
                gtk.STOCK_DISCONNECT)
            optionsMenu.add(pluginMenuItem)
        
        preferencesMenuItem = self.newStockImageMenuItem(
            gtk.STOCK_PREFERENCES)
        preferencesMenuItem.add_accelerator('activate', accelGroup, \
            ord('P'), gtk.gdk.CONTROL_MASK, gtk.ACCEL_VISIBLE)
        optionsMenu.add(preferencesMenuItem)

        accelGroup.connect_group(ord('P'), gtk.gdk.CONTROL_MASK, \
            gtk.ACCEL_LOCKED, self.on_preferences_activate)

        # ---------------------- HELP MENU --------------------------------

        aboutMenuItem = self.newStockImageMenuItem(gtk.STOCK_ABOUT)
        helpMenu.add(aboutMenuItem)

        self.add(fileMenuItem)

        # No Actions Menu in login window
        if window_type == 'userlist':
            self.add(viewMenuItem)
            self.add(actionsMenuItem)

        self.add(optionsMenuItem)
        self.add(helpMenuItem)

        # -------------------- CALLBACKS ----------------------------------

        quitMenuItem.connect('activate', self.on_quit_activate)
        aboutMenuItem.connect('activate', self.on_about_activate)
        # We don't need this callbacks in login window, 
        # because we didn't create this menu items
        if window_type == 'userlist':
            disconnect.connect('activate', self.on_logout_activate)
            self.orderByGroup.connect('activate', self.on_order_changed)
            self.orderByStatus.connect('activate', self.on_order_changed)
            self.showByNick.connect('activate', 
                self.on_show_by_nick_activate)
            self.showOffline.connect('activate', 
                self.on_show_offline_activate)
            self.showEmptyGroups.connect('activate', 
                self.on_show_empty_groups_activate)
            self.showCountContact.connect('activate', 
                self.on_show_count_contact_activate)
            addUserMenuItem.connect('activate', self.on_add_user_activate)
            setAliasAction.connect('activate', self.on_rename_user_activate)
            deleteAction.connect('activate', self.on_delete_user_activate)
            blockAction.connect('activate', self.on_block_user_activate)
            unblockAction.connect('activate', self.on_unblock_user_activate)
            addGroupMenuItem.connect('activate', self.on_add_group_activate)
            deleteGAction.connect('activate', self.on_delete_group_activate)
            renameAction.connect('activate', self.on_rename_group_activate)
            setAutoReplyMenuItem.connect('activate', 
                self.on_set_auto_reply_activate)
            self.activateAutoReply.connect('activate', 
                self.on_auto_reply_activate)
            changeNickMenuItem.connect('activate', 
                self.on_change_nick_activate)
            changeAvatarMenuItem.connect('activate', 
                self.on_change_avatar_activate)
            pluginMenuItem.connect('activate', self.on_plugin_activate)
        else:
            getliveMenuItem.connect('activate', self.on_get_live_activate)
            
        preferencesMenuItem.connect('activate', 
            self.on_preferences_activate)

        # ----------------- VARS ---------------------------

    def newStockImageMenuItem (self, stock):
        '''create a new image menu item from gtk's stock and retrun it'''

        mi = gtk.ImageMenuItem(stock)
        return mi

    def newImageMenuItem(self, label, stock = None, img = None):
        '''create a new Imege menu item and return it, it could have a 
            stock image or a custom image'''

        mi = gtk.ImageMenuItem(_(label))

        if stock != None:
            mi.set_image(gtk.image_new_from_stock(stock, 
                gtk.ICON_SIZE_MENU))
        elif img != None:
            image = gtk.Image()
            image.set_from_pixbuf(img)
            mi.set_image(image)

        return mi

    def newCheckMenuItem(self, label, checked):
        '''create a new checkbox and return it, if checked is true, 
            the check box will be checked (d'uh!)'''

        mi = gtk.CheckMenuItem(_(label))
        mi.set_active(checked)
        return mi

    # -------------------------- CALLBACKS -------------------------------

    def on_move_to_activate(self, menuItem, group):
        self.controller.contacts.move_to_group(self.userName, 
            self.userGroup, group)

    def on_rename_user_activate(self, *args):
        self.controller.contacts.set_alias_dialog(self.userName)

    def on_actions_activate(self, *args):
        data = self.controller.getMenuData()
        typeSelected = data[0]

        self.contactsActionGroup.set_sensitive(typeSelected == 'user')
        self.groupsActionGroup.set_sensitive\
                    (typeSelected == 'group' and data[2] != 'nogroup')

        # set selected data class-available
        # set some action's senstibility
        if typeSelected == 'user':
            self.userName = data[1]
            self.userGroup = data[2]
            # block/unblock
            blocked = self.controller.contacts.get_blocked(data[1])
            self.blockAction.set_sensitive(not blocked)
            self.unblockAction.set_sensitive(blocked)
        if typeSelected == 'group':
            self.groupName = data[1]

    def on_quit_activate(self, *args):
        self.controller.quit(0)

    def on_preferences_activate(self, *args):
        if not self.controller.preference_open:
            PreferenceWindow.PreferenceWindow(self.controller, 
                self.config, self.controller.mainWindow).show()
            self.controller.preference_open = True


    def on_about_activate(self, *args):
        try:
            f = file('COPYING', 'r')
        except:
            f = None

        def closeAbout(widget, response_id):
            if response_id == gtk.RESPONSE_CANCEL:
                widget.destroy()

        gtk.about_dialog_set_url_hook(self.on_click_url, None)
        about = gtk.AboutDialog()
        about.set_name('emesene')
        about.set_version('1.0.1')
        about.set_copyright('Luis Mariano Guerra')
        about.set_comments(_('A client for the WLM%s network') %'\xe2\x84\xa2')
        about.connect('response', closeAbout)

        if f == None:
            about.set_license('GNU General Public License')
        else:
            about.set_license(f.read())

        about.set_website('http://www.emesene.org')

        about.set_authors([
            'Luis Mariano Guerra (emesene and emesenelib)', \
            'Horacio Duran (emesene and emesenelib)', \
            'Alberto Talavera (emesene)', \
            'Linan Wang (MsnOIM)', \
            'Roberto Salas & Jakub Steiner (tango theme)', \
            'Vinicius Depizzol (default theme)', \
            'Yguaratã C. Cavalcanti (emesene)', \
            'Roger Duran (emesene)', \
            'dx (emesene & emesenelib)', \
            'alencool (emesene & cairo wizzard :P)', \
            'Mattia \'MaTz\' Pini (emesene)', \
            'mg (emesene)', \
            'Jan de Mooij (emesene)', \
            'j0hn (emesene)', \
            'Luis \'JoinTheHell\' Nell (emesene)', \
            'nopersona (in the margins theme)'
            ])
        about.set_translator_credits(_('translator-credits'))
        icon = self.controller.theme.getImage('login')
        about.set_icon(icon)
        about.set_logo(icon)
        about.run()
    
    def on_click_url(self, dialog, link, user_data):
        desktop.open(link)

    def on_logout_activate(self, *args):
        self.controller.logout()
            
    def on_order_changed(self, menuitem):
        if menuitem == self.orderByGroup:
            self.controller.config.user['orderByStatus'] = False
        elif menuitem == self.orderByStatus:
            self.controller.config.user['orderByStatus'] = True
    
        self.controller.refreshUserList(force=True)

    def on_show_by_nick_activate(self, *args):
        self.controller.config.user['showByNick'] = \
            self.showByNick.get_active()
        self.controller.refreshUserList()

    def on_show_offline_activate(self, *args):
        self.controller.config.user['showOffline'] = \
            self.showOffline.get_active()
        self.controller.mainWindow.userList.refilter()

    def on_show_empty_groups_activate(self, *args):
        self.controller.config.user['showEmptyGroups'] = \
            self.showEmptyGroups.get_active()
        self.controller.mainWindow.userList.refilter()
    
    def on_show_count_contact_activate(self, *args):
        self.controller.config.user['showCountContact'] = \
            self.showCountContact.get_active()
        self.controller.refreshUserList(force=True)

    def on_add_user_activate(self, *args):
        self.controller.addUserDialog()

    def on_delete_user_activate(self, *args):
        self.controller.contacts.remove(self.userName)

    def on_block_user_activate(self, *args):
        self.controller.contacts.block(self.userName)

    def on_unblock_user_activate(self, *args):
        self.controller.contacts.unblock(self.userName)

    def on_add_group_activate(self, *args):
        self.controller.groups.add_dialog()

    def on_delete_group_activate(self, *args):
        self.controller.groups.remove(self.groupName)

    def on_rename_group_activate(self, *args):
        self.controller.groups.rename_dialog(self.groupName)

    def on_change_avatar_activate(self, *args):
        self.controller.set_picture_dialog()

    def on_change_nick_activate(self, *args):
        self.controller.contacts.set_nick_dialog()

    def on_set_auto_reply_activate(self, *args):
        def response_cb(response, message=''):
            '''callback for the set autoreply dialog'''
            if response == stock.ACCEPT:
                if message == '':
                    dialog.error(_("Empty autoreply"))
                else:
                    self.controller.autoReplyMessage = message 
                    self.config.user['autoReplyMessage'] = message
                
        window = dialog.entry_window(_("Autoreply message:"), 
            self.config.user['autoReplyMessage'], response_cb, 
            _("Change autoreply"))
        window.show()

    def on_auto_reply_activate(self, *args):
        self.config.user['autoReply'] = self.activateAutoReply.get_active()

    def on_plugin_activate(self, *args):
        self.controller.pluginDialog()
    
    def on_get_live_activate(self, *args):
        link = 'https://accountservices.passport.net/reg.srf?sl=1'
        desktop.open(link)




# -*- coding: utf-8 -*-

#   This file is part of emesene.
#
#    Emesene is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    emesene is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with emesene; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import gtk
import cairo
try:
    import Theme
    import Config
except:
    Theme = None

WIDTH=64
HEIGHT=64

class ImageAreaSelectorDialog(gtk.Dialog):
    def __init__(self, pixbuf, title = "Select area of image", parent = None):
        gtk.Dialog.__init__(self, title, parent,
                            gtk.DIALOG_DESTROY_WITH_PARENT,
                            (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL,
                             gtk.STOCK_OK, gtk.RESPONSE_OK ))
        
        
        self.set_default_response(gtk.RESPONSE_CANCEL)
        self.set_resizable(False)
        self.selector = ImageAreaSelector()
        self.selector.set_from_pixbuf(pixbuf)
        bbox = gtk.HButtonBox()
        self.button_rcw = gtk.Button(label="Rotate")
        self.button_rcw.connect("clicked", self._on_rcw_clicked)
        self.button_rccw = gtk.Button(label="Rotate")
        self.button_rccw.connect("clicked", self._on_rccw_clicked)

        if Theme:
            theme = Theme.Theme(Config.Config())  # FIXME ARRGH
            image_90 = gtk.Image()
            image_90.set_from_pixbuf(theme.getImage("rotate-90"))
            self.button_rcw.set_image(image_90)
            image_270 = gtk.Image()
            image_270.set_from_pixbuf(theme.getImage("rotate-270"))
            self.button_rccw.set_image(image_270)
        

        if gtk.gtk_version >= (2, 10, 0):
            bbox.pack_start(self.button_rccw)
            bbox.pack_end(self.button_rcw)
        self.vbox.pack_start(self.selector)
        self.vbox.pack_end(bbox)
        self.vbox.show_all()

    def run(self):
        response = gtk.Dialog.run(self)
        if response == gtk.RESPONSE_OK:
            pixbuf = self.selector.get_selected()
            self.destroy()
            return (response, pixbuf)
        else:
            self.destroy()
            return (gtk.RESPONSE_CANCEL, None)

    def _on_rcw_clicked(self, *args):
        self.selector.rotate(1)
    def _on_rccw_clicked(self, *args):
        self.selector.rotate(-1)

class ImageAreaSelector(gtk.DrawingArea):
    __gtype_name__ = "ImageAreaSelector"
    def __init__(self):
        gtk.DrawingArea.__init__(self)
        self.connect("expose_event", self.expose)
        self.image = None
        self.pixbuf = None
        self.connect("button_press_event", self.button_press)
        self.connect("button_release_event", self.button_release)
        self.connect("motion_notify_event", self.motion_notify)
        self.connect("configure_event", self.configure_event)
        self.add_events(gtk.gdk.BUTTON_PRESS_MASK |
                        gtk.gdk.BUTTON_RELEASE_MASK |
                        gtk.gdk.POINTER_MOTION_MASK) 

        self._pressed = False
        self._moved = False
        self._state = gtk.gdk.LEFT_PTR
        self._has_overlay = False
        self._angle = 0

    def get_selected(self):
        if gtk.gtk_version >= (2, 10, 0):
            angle = self.get_angle()
            self._trans_pixbuf = self._trans_pixbuf.rotate_simple(angle)
        return self._trans_pixbuf.subpixbuf(*self.selection)
        
    def _get_selection(self):
        return (self._x, self._y, self._width, self._height)
    def _set_selection(self, (x, y, width, height)):
        self._x = x
        self._y = y
        self._width = width
        self._height = height
        self.redraw_canvas()
    selection = property(_get_selection, _set_selection)

    def _get_sel_coord(self):
        x1 = self._x
        y1 = self._y
        x2 = x1 + self._width
        y2 = y1 + self._height
        return (x1, y1, x2, y2)

    def _set_sel_coord(self, x1, y1, x2, y2):
        if x1 < 0 or y1 < 0 or x2 > self.pixbuf.get_width() \
               or y2 > self.pixbuf.get_height():
            return
        if x1 > x2:
            x2 = x1
        if y1 > y2:
            y2 = y1
            
        self.selection = map(int, (x1, y1, x2-x1, y2-y1))

    def reset_selection(self):
        self.selection = (0,0,0,0)

    def button_release(self, *args):
        self._pressed = False
        if not self._moved and self._state == gtk.gdk.LEFT_PTR:
            self.reset_selection()


    def expose(self, widget, event):
        if not self._has_overlay:
            self.create_overlay()
        x , y, width, height = event.area

        widget.window.draw_drawable(widget.get_style().fg_gc[gtk.STATE_NORMAL],
                                    self.pixmap_shaded, 0, 0,
                                    0, 0,
                                    self.pixbuf.get_width(),
                                    self.pixbuf.get_height())

        widget.window.draw_pixbuf(None, pixbuf=self.pixbuf,
                                  src_x=self._x,
                                  src_y=self._y,
                                  dest_x= self._x,
                                  dest_y=self._y ,
                                  width=self._width, height=self._height,
                                  dither=gtk.gdk.RGB_DITHER_NORMAL,
                                  x_dither=0, y_dither=0)
        return False


        
    def rotate(self, angle):
        self._angle = (self._angle + angle) % 4
        angle = self.get_angle()
        self._init_pixbuf(angle, False)
    
    def get_angle(self):
        if not gtk.gtk_version >= (2, 10, 0):
            return 0

        if self._angle == 1:
            angle = gtk.gdk.PIXBUF_ROTATE_CLOCKWISE
        elif self._angle == 2:
            angle = 180
        elif self._angle == 3:
            angle = gtk.gdk.PIXBUF_ROTATE_COUNTERCLOCKWISE
        else:
            angle = 0    
        return angle
        
    def update_selection(self, event):
        # track mouse delta
        dx = event.x - self._mouse_x
        dy = event.y - self._mouse_y
            
        x1, y1, x2, y2 = self._get_sel_coord()
        if self._state == gtk.gdk.TOP_LEFT_CORNER:
            w = x2 - event.x
            h = y2 - event.y
            delta = max((w, h))
            x1 = x2 - delta
            y1 = y2 - delta
        elif self._state == gtk.gdk.TOP_RIGHT_CORNER:
            w = event.x - x1
            h = y2 - event.y
            delta = max((w, h))
            x2 = x1 + delta
            y1 = y2 - delta
        elif self._state == gtk.gdk.BOTTOM_RIGHT_CORNER:
            w = (event.x-x1)
            h = (event.y-y1)
            delta = max((w, h))
            x2 = x1 + delta
            y2 = y1 + delta
        elif self._state == gtk.gdk.BOTTOM_LEFT_CORNER:
            w = x2 - event.x
            h = (event.y-y1)
            delta = max((w, h))
            x1 = x2 - delta
            y2 = y1 + delta
        elif self._state == gtk.gdk.FLEUR:
            x1 += dx
            y1 += dy
            x2 += dx
            y2 += dy
        else:
            # mouse pressed outside current selection. Create new selection
            x1 = self._press_x
            y1 = self._press_y
            w = event.x - x1
            h = event.y - y1
            delta = max((w, h))
            x2 = x1 + delta
            y2 = y1 + delta
        
        self._mouse_x = event.x
        self._mouse_y = event.y
        self._set_sel_coord(x1, y1, x2, y2)

    def update_cursor(self, event):
        fuzz = max((5, int(self._width * 0.05)))
        if abs(event.y - self._y) < fuzz:
            if abs(event.x -self._x) < fuzz:
                self._state = gtk.gdk.TOP_LEFT_CORNER
            elif abs(event.x - (self._x + self._width)) < fuzz:
                self._state = gtk.gdk.TOP_RIGHT_CORNER
        elif abs(event.y - (self._y + self._height)) < fuzz:
            if abs(event.x - self._x) < fuzz:
                self._state = gtk.gdk.BOTTOM_LEFT_CORNER
            elif abs(event.x -(self._x + self._width)) < fuzz:
                self._state = gtk.gdk.BOTTOM_RIGHT_CORNER
        elif event.x > self._x and event.x < self._x + self._width \
                 and event.y > self._y and event.y < self._y + self._height:
            self._state = gtk.gdk.FLEUR
        else:
            self._state = gtk.gdk.LEFT_PTR

        self.window.set_cursor(gtk.gdk.Cursor(self._state))
        
    def motion_notify(self, widget, event):
        if self._pressed:
            self._moved = True
            self.update_selection(event)
        else:
            self.update_cursor(event)
            
    def button_press(self, widget, event):
        self._pressed = True
        self._moved = False
        self._mouse_x = event.x
        self._mouse_y = event.y
        self._press_x = event.x
        self._press_y = event.y
        if event.button == 3:
            self.reset_selection()

    def configure_event(self, widget, event):
        x, y, width, height = widget.get_allocation()

        self.pixmap = gtk.gdk.Pixmap(widget.window, width, height)
        self.pixmap.draw_rectangle(widget.get_style().white_gc,
                                   True, 0, 0, width, height)

        self.pixmap_shaded = gtk.gdk.Pixmap(widget.window, width, height)
        self.pixmap_shaded.draw_rectangle(widget.get_style().white_gc,
                                          True, 0, 0, width, height)
        self._has_overlay = False
        return True
                          
    def create_overlay(self):
        context = self.pixmap_shaded.cairo_create()
        width = self.pixbuf.get_width()
        height = self.pixbuf.get_height()

        target  = context.get_target()
        overlay = target.create_similar(cairo.CONTENT_COLOR_ALPHA, width, height)
        punch   = target.create_similar(cairo.CONTENT_ALPHA, width, height)
        context.set_source_pixbuf(self.pixbuf, 0, 0)
        context.fill()
        context.paint()

        # Draw a grey overlay
        overlay_cr = cairo.Context (overlay)
        overlay_cr.set_source_rgba (0.4, 0.4, 0.4, 0.6)
        overlay_cr.rectangle(0, 0, width, height)
        overlay_cr.fill()

        context.set_source_surface (overlay, 0, 0)
        context.paint()
        self._has_overlay = True


    def redraw_canvas(self):
        if self.window:
            alloc = self.get_allocation()
            self.queue_draw_area(alloc.x, alloc.y, alloc.width, alloc.height)

    def do_size_request(self, requisition):
        if not self.pixbuf:
            return
        requisition.width = self.pixbuf.get_width()
        requisition.height = self.pixbuf.get_height()
        
    def set_from_pixbuf(self, pixbuf):
        h = pixbuf.get_height()
        w = pixbuf.get_width()
        edge = max(w, h)
        self._trans_pixbuf = gtk.gdk.Pixbuf(gtk.gdk.COLORSPACE_RGB, True, 8,
                                            edge, edge)
        self._disp_pixbuf = gtk.gdk.Pixbuf(gtk.gdk.COLORSPACE_RGB, True, 8,
                                           edge, edge)
        
        self._trans_pixbuf.fill(0xffffff00)
        self._disp_pixbuf.fill(0x66666666)
        
        dx = (edge-w)/2
        dy = (edge-h)/2
        
        pixbuf.composite(self._trans_pixbuf, dx, dy, w, h, dx, dy, 1, 1,
                         gtk.gdk.INTERP_BILINEAR, 255)
        pixbuf.composite(self._disp_pixbuf, dx, dy, w, h, dx, dy, 1, 1,
                         gtk.gdk.INTERP_BILINEAR, 255)

        maxw = int(0.75 * gtk.gdk.screen_width())
        maxh = int(0.75 * gtk.gdk.screen_height())
        if w > maxw or h > maxh:
            wscale = float(maxw) / edge
            hscale = float(maxh) / edge
            scale = min(wscale, hscale)
            
            edge = int(edge * scale)
            
            self._trans_pixbuf = self._trans_pixbuf.scale_simple(edge, edge,
                                                    gtk.gdk.INTERP_BILINEAR )
            self._disp_pixbuf = self._disp_pixbuf.scale_simple(edge, edge,
                                                    gtk.gdk.INTERP_BILINEAR )

        self._init_pixbuf()

    def _init_pixbuf(self, angle=None, create_selection=True):
        self.pixbuf = self._disp_pixbuf.copy()

        if angle and gtk.gtk_version >= (2, 10, 0):
            self.pixbuf = self.pixbuf.rotate_simple(angle)
            self.configure_event(self, None)

        self._has_overlay = False
        w = self.pixbuf.get_width()
        h = self.pixbuf.get_height()
        
        if create_selection:
            sw = min((w, h))
            x1 = int((w-sw)/2)
            y1 = int((h-sw)/2)
            self.selection = (x1, y1, sw, sw)
        else:
            self.selection = (0, 0, 0, 0)                   



# -*- coding: utf-8 -*-

#   This file is part of emesene.
#
#    Emesene is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    emesene is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with emesene; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
'''a class that represent a contextual menu of a group'''

import Menu
import stock

class GroupMenu(Menu.Menu):
    '''class that represent a contextual menu of a group'''

    def __init__(self, group, groups, contacts, dialog):
        '''class constructor'''
        Menu.Menu.__init__(self)

        self.group = group
        self.groups = groups
        self.contacts = contacts
        self.dialog = dialog
        
        self._build()

    def _build(self):
        '''build the menu'''
        self.remove_item = Menu.Item(_('_Remove'), Menu.Item.TYPE_STOCK, 
            stock.REMOVE)
        self.remove_item.signal_connect('selected', self._on_remove_selected)
        
        self.rename_item = Menu.Item(_('Re_name'), Menu.Item.TYPE_STOCK, 
            stock.EDIT)
        self.rename_item.signal_connect('selected', self._on_rename_selected)

        self.add_item = Menu.Item(_('_Add contact'), Menu.Item.TYPE_STOCK, 
            stock.ADD)
        self.add_item.signal_connect('selected', self._on_add_selected)

        self.append(self.remove_item)
        self.append(self.rename_item)
        self.append(self.add_item)
    
    def _on_remove_selected(self, item):
        '''called when remove is selected'''
        def _yes_no_cb(response):
            '''callback from the confirmation dialog'''
            if response == stock.YES:
                self.groups.remove(self.groups.name)

        self.dialog.yes_no(
            _('Are you sure you want to remove group %s?') % \
            (self.group.name,), _yes_no_cb)

    def _on_rename_selected(self, item):
        '''called when rename is selected'''
        self.groups.rename_dialog(self.group.name)
    
    def _on_add_selected(self, item):
        '''called when rename is selected'''
        def add_contact_cb(response, account='', group=''):
            '''callback for the add_contact dialog'''
            if response == stock.ACCEPT:
                if account:
                    self.contacts.add(account)

                    if group:
                        self.contacts.add_to_group(account, group)
                else:
                    self.dialog.warning(_("Empty account"))

        self.dialog.add_contact(self.groups.groups.keys(), self.group.name, 
            add_contact_cb)




# -*- coding: utf-8 -*-

#   This file is part of emesene.
#
#    Emesene is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    emesene is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with emesene; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

VERSION = '0.1'

import Plugin
from Parser import Url
import gtk
import urllib

ERROR = ''
MIMETYPE = 'application/x-youtube-thumb'
OBJECT_TEMPLATE = '<object class="%s" type="' + MIMETYPE + '"/>'
THUMBNAIL_URL = 'http://img.youtube.com/vi/%s/default.jpg'

class MainClass( Plugin.Plugin ):
    '''Main plugin class'''
    
    def __init__( self, controller, msn ):
        '''Constructor'''
        
        Plugin.Plugin.__init__( self, controller, msn )
        
        self.description = _( 'Displays thumbnails of youtube videos next to links' )
        self.authors = { 'Roger Duran' : 'RogerDuran at gmail dot com' }
        self.website = 'http://www.rogerpc.com.ar'
        self.displayName = _( 'Youtube' )
        self.name = 'Youtube'

        self.controller = controller
        self.parser = controller.unifiedParser
        self.convManager = controller.conversationManager
        self.config = controller.config

        self.postParseId = None

    def start( self ):
        '''start the plugin'''
        # this signals should be emitted by conversationManager
        self.postParseId = self.parser.connect('filter', self.parseUrl )
        self.convManager.connect( 'new-conversation-ui', \
            self.newConversationWindow )

        for conversation in self.getOpenConversations():
            textView = conversation.ui.textview
            textView.customObjectsCallbacks.update({
                MIMETYPE: (self.youtubeWidget,None)
             })
        self.enabled = True
        
    def stop( self ):    
        '''stop the plugin'''
        self.controller.unifiedParser.disconnect( self.postParseId )
        for conversation in self.getOpenConversations():
            textView = conversation.ui.textview
            if MIMETYPE in textView.customObjectsCallbacks:
                textView.customObjectsCallbacks.pop(MIMETYPE)
        self.enabled = False
        
    def action( self ):
        pass
    
    def check( self ):
        '''
        check if everything is OK to start the plugin
        return a tuple whith a boolean and a message
        if OK -> ( True , 'some message' )
        else -> ( False , 'error message' )
        '''
        if ERROR != '':
            return ( False, ERROR )
            
        return ( True, 'Ok' )
        

    def newConversationWindow( self, conversationmanager, conversation, win ):
        textView = conversation.ui.textview
        textView.customObjectsCallbacks.update({
                       MIMETYPE: (self.youtubeWidget,None)
                     })

    def youtubeWidget( self, textview, id, path=None ):
        img = gtk.Image()
        img.set_from_stock( gtk.STOCK_MISSING_IMAGE, \
            gtk.ICON_SIZE_SMALL_TOOLBAR )
        img.show()
        def getImage():
            try:
                mem = urllib.urlopen(THUMBNAIL_URL % id).read(10*1024*1024)
                loader = gtk.gdk.PixbufLoader()
                loader.write(mem)
                loader.close()
                return (textview, loader, img)
            except:
                return
        defrun = Plugin.Function(getImage, self.callback)
        defrun()
        return img
        
    def callback(self, handler):
        try:
            textview, loader, img = handler
            textview.scrollLater()
            # FIXME: do we need animation?
            img.set_from_animation(loader.get_animation())
        except:
            pass

    def parseUrl( self, parser, dataType, filterdata ):
        if not dataType or dataType.getDataType() != 'conversation':
            return

        newObjects = []
        for object in filterdata.list:
            code = ''
            if type(object) == Url and \
               object.url.count('youtube.com') and \
               object.url.count('watch?v='):
                
                videoid = object.url.split('watch?v=')[1].split('&')[0]
                code = OBJECT_TEMPLATE % videoid
                html = object.getHtml()
                object.getHtml = lambda: code + html

                
            newObjects.append(object)
        filterdata.list = newObjects



# -*- coding: utf-8 -*-

#   This file is part of emesene.
#
#    Emesene is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    emesene is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with emesene; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import gtk
import gtk.gdk
import cairo
import gobject

class FancyAvatarRenderer ( gtk.GenericCellRenderer ):
    """Renderer for avatar """
    
    __gproperties__ = {
        'pixbuf': (gtk.gdk.Pixbuf, 'Pixbuf','Pixbuf',gobject.PARAM_READWRITE),        
        'blocked': (bool , 'User Blocked status', '',False, gobject.PARAM_READWRITE),        
        'dimention' : (gobject.TYPE_INT, 'cell dimentions', 'height width of cell', 0, 96, 32, gobject.PARAM_READWRITE),
        'status': (str, 'Contact status', '','FLN', gobject.PARAM_READWRITE),
        'radius_factor' : (gobject.TYPE_FLOAT,'radius of pixbuf', 
        '0.0 to 0.5 with 0.1 = 10% of dimention',0.0, 0.5,0.11, gobject.PARAM_READWRITE),
         }         
        
    def __init__(self, controller, cellDimention = 32, cellRadius = 0.11):
        #gobject.GObject.__init__(self)
        self.__gobject_init__()
        self._pixbuf = None
        self._status = 'FLN'
        self._blocked = False
        self._dimention = cellDimention
        self._radius_factor = cellRadius
        
        self.set_property('xpad', 1)
        self.set_property('ypad', 1)   
        
        self._theme = controller.theme
        self._config = controller.config
        
        #set up information of statusTransformation
        self.__set_transformation(self._config.user['statusTransformation'])

        self.transId = self._config.connect('change::statusTransformation', \
            self.__transformation_callback)
        
    def destroy(self):
        self._config.disconnect(self.transId)
        gtk.GenericCellRenderer.destroy(self)

    def __set_transformation(self, setting):
        transformation = setting.split('|')                   
        if 'pixelate' in transformation: self._pixalated = True  
        else: self._pixalated = False
        if 'corner' in transformation: self._corner = True  
        else: self._corner = False        
        if 'alpha' in transformation: self._alpha_status = True  
        else: self._alpha_status = False
        if 'mini' in transformation: self._mini = True  
        else: self._mini = False

    def __transformation_callback(self, config, newvalue, oldvalue):
        self.__set_transformation(newvalue)
        
    def do_get_property(self, property):
        if property.name == 'pixbuf':
            return self._pixbuf
        elif property.name == 'dimention':
            return self._dimention
        elif property.name == 'radius-factor':
            return self._radius_factor
        elif property.name == 'blocked':
            return self._blocked
        elif property.name == 'status':
            return self._status                        
        else:
            raise AttributeError, 'unknown property %s' % property.name
  
    def do_set_property(self, property, value):
        if property.name == 'pixbuf':
            self._pixbuf = value
        elif property.name == 'dimention':            
            self._dimention = value
        elif property.name == 'radius-factor':
            self._radius_factor = value         
        elif property.name == 'blocked':
            self._blocked = value
        elif property.name == 'status':
            self._status = value      
        else:
            raise AttributeError, 'unknown property %s' % property.name

    def on_get_size( self, widget, cell_area=None ):       
        xpad,ypad = self.get_property('xpad'), self.get_property('ypad')
        if self._dimention >= 32: width = self._dimention
        elif self._mini: width = self._dimention
        elif self._corner: width = self._dimention * 2            
        else: width = self._dimention        
        height = self._dimention + (ypad *2)                       
        return ( 0, 0,  width, height) 
        
    def on_render( self, window, widget, background_area, cell_area, expose_area, flags ):        
        avatar = self._pixbuf
        overlay = None
        alpha = 1
                 
        if self._pixalated: avatar = __getPixalate(self._pixbuf)            
        if self._corner: overlay = self.__getOverlay()
        if self._alpha_status: alpha = self.__getAlpha()
        
        xpad,ypad = self.get_property('xpad'), self.get_property('ypad')
        cell_x, cell_y, cell_width, cell_height = cell_area 
        ctx = window.cairo_create()          
        ctx.translate(cell_x,cell_y)
        
        if not avatar == None:
            self.__drawScalePixbuf( ctx,avatar,cell_width - self._dimention,ypad ,self._dimention, 
                gtk.ANCHOR_CENTER, self._radius_factor, alpha)
        
        if not overlay == None:       
            if self._dimention >= 32 :     
                self.__drawScalePixbuf( ctx,overlay,cell_width - 16, ypad + self._dimention - 16,16, position = gtk.ANCHOR_SW)
            elif self._mini:
                self.__drawScalePixbuf( ctx,overlay,cell_width - 8, ypad + self._dimention - 8,8, position = gtk.ANCHOR_SW)
            else:
                self.__drawScalePixbuf( ctx,overlay,0,ypad  ,16)
             
    def __getAlpha(self):
        if self._status in ['IDL', 'FLN']: return 0.75
        else: return 1

    def __getOverlay(self):
        if self._blocked:
            return self._theme.getImage('status-blocked')
        elif self._status in ['AWY','BRB','LUN','IDL' ] :
            return self._theme.getImage('status-away')
        elif self._status in ['BSY','PHN' ] :
            return self._theme.getImage('status-busy')
        else:
            return None       
            
    def __getPixalate( self, pixbuf ):
        pixbuf_to_modify = pixbuf.copy()        
        if not self._status == 'NLN':
            if contact.status == 'BSY':
                pixbuf_to_modify.saturate_and_pixelate(pixbuf,1.0, True)
            else:
                pixbuf_to_modify.saturate_and_pixelate(pixbuf,0.1, False)                
        return pixbuf_to_modify
        
          
    def __drawScalePixbuf(self, ctx, pixbuf, x,y , dimention, position = gtk.ANCHOR_CENTER, radius = 0, alpha = 1 ):
        
        ctx.save()
        ctx.set_antialias(cairo.ANTIALIAS_SUBPIXEL)
        ctx.translate(x,y)
        
        pix_width = pixbuf.get_width()
        pix_height = pixbuf.get_height()
        if (pix_width > dimention) or (pix_height > dimention): 
            scale_factor = float(dimention) / max (pix_width,pix_height)
        else: 
            scale_factor = 1        
        scale_width = pix_width* scale_factor
        scale_height = pix_height* scale_factor 
 
        #tranlate position
        if position in [gtk.ANCHOR_NW,gtk.ANCHOR_W,gtk.ANCHOR_SW] : x = 0
        elif position in [gtk.ANCHOR_N,gtk.ANCHOR_CENTER,gtk.ANCHOR_S] : x = (dimention/2) - (scale_width/2)
        else: x = dimention - scale_width        
        if position in [gtk.ANCHOR_NW,gtk.ANCHOR_N,gtk.ANCHOR_NE] : y = 0
        elif position in [gtk.ANCHOR_E,gtk.ANCHOR_CENTER,gtk.ANCHOR_W] : y = (dimention/2) - (scale_height/2)
        else: y = dimention - scale_height         
        ctx.translate(x, y)
            
        if radius > 0 : 
            self.__roundedrecMoonlight(ctx,0,0,scale_width,scale_height, self._dimention * radius) 
            ctx.clip()
        ctx.scale(scale_factor,scale_factor)
        ctx.set_source_pixbuf(pixbuf,0,0)
        ctx.paint_with_alpha(alpha)
        ctx.restore()
    
    def __roundedrecMoonlight(self, cr,x,y,w,h,radius=5):
        # http://cairographics.org/cookbook/roundedrectangles/
        # modified from mono moonlight aka mono silverlight
        # test limits (without using multiplications)
        # http://graphics.stanford.edu/courses/cs248-98-fall/Final/q1.html

        ARC_TO_BEZIER = 0.55228475
        if radius > (min(w,h)/2):
            radius = (min(w,h)/2)
        #approximate (quite close) the arc using a bezier curve
        c = ARC_TO_BEZIER * radius

        cr.new_path();
        cr.move_to ( x + radius, y)
        cr.rel_line_to ( w - 2 * radius, 0.0)
        cr.rel_curve_to ( c, 0.0, radius, c, radius, radius)
        cr.rel_line_to ( 0, h - 2 * radius)
        cr.rel_curve_to ( 0.0, c, c - radius, radius, -radius, radius)
        cr.rel_line_to ( -w + 2 * radius, 0)
        cr.rel_curve_to ( -c, 0, -radius, -c, -radius, -radius)
        cr.rel_line_to (0, -h + 2 * radius)
        cr.rel_curve_to (0.0, -c, radius - c, -radius, radius, -radius)
        cr.close_path ()
                        
gobject.type_register( FancyAvatarRenderer )



# -*- coding: utf-8 -*-

#   This file is part of emesene.
#
#    Emesene is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    emesene is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with emesene; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import gtk
import pango
import os

import ConversationUI
import emesenelib.common  # -.-"

class ConversationWindow(gtk.Window):
    '''This class is the conversation window, it is just the frame
    where all the conversations are hold, it can be tabbed or one
    conversation per window, that depend on the config.'''

    def __init__(self, controller, conversation):
        '''Constructor'''
        gtk.Window.__init__(self)

        self.set_border_width(0)

        self.controller = controller
        self.config = controller.config
        self.closed = False
        self.notOpen = True

        self.width = self.config.user['convWindowWidth']
        self.height = self.config.user['convWindowHeight']
        self.set_default_size(self.width, self.height)
        self.connect('size-allocate', self.on_size_alloc)
        
        # accelerators
        accelGroup = gtk.AccelGroup()
        self.add_accel_group(accelGroup)
        self.accelGroup = accelGroup
        accelGroup.connect_group(gtk.keysyms.Page_Down, \
                                  gtk.gdk.CONTROL_MASK, gtk.ACCEL_LOCKED, \
                                  self.on_key_cycle_tabs)
        accelGroup.connect_group(gtk.keysyms.Page_Up, \
                                  gtk.gdk.CONTROL_MASK, gtk.ACCEL_LOCKED, \
                                  self.on_key_cycle_tabs)
        accelGroup.connect_group(gtk.keysyms.W, \
                                  gtk.gdk.CONTROL_MASK, gtk.ACCEL_LOCKED, \
                                  self.on_key_close_tab)
        
        for i in range(1, 10):
            accelGroup.connect_group(gtk.keysyms._0 + i, \
                                  gtk.gdk.MOD1_MASK, gtk.ACCEL_LOCKED, \
                                  self.on_key_change_tab)
        
        self.conversation = conversation
        self.conversation.setIsCurrent(True)
        
        self.update_title()
        self.set_role('chat')
        
        self.textTagCount = 0

        self.user = self.conversation.getUser()
        self.conversations = []

        self.menu = ConversationWindowMenu(self, controller)
        
        self.vbox = gtk.VBox()
        self.tabs = gtk.Notebook()
        self.tabs.set_show_border(False)
        self.tabs.set_property('tab-vborder', 0)

        self.tabs.set_show_tabs(False) #ok, ok, never show tabs on window init
        self.tabs.set_scrollable(True)
        self.tabs.show_all()
        self.openTab(conversation)

        self.vbox.pack_start(self.menu, False, False)
        self.vbox.pack_start(self.tabs, True, True)
        
        if self.config.user['showMenubar']:
            self.menu.show_all()
        else:
            accelGroup.connect_group(ord('L'), gtk.gdk.CONTROL_MASK, \
                gtk.ACCEL_LOCKED, self.clearOutputText)
            accelGroup.connect_group(ord('Q'), gtk.gdk.CONTROL_MASK, \
                gtk.ACCEL_LOCKED, self.menu.onCloseActivate)
            accelGroup.connect_group(ord('I'), gtk.gdk.CONTROL_MASK, \
                gtk.ACCEL_LOCKED, self.menu.onInviteActivate)
            accelGroup.connect_group(ord('S'), gtk.gdk.CONTROL_MASK, \
                gtk.ACCEL_LOCKED, self.menu.onSendFileActivate)
        
        self.vbox.show()

        self.connect('delete-event', self.close)
        self.connect('focus-in-event', self.on_focus_in_event)
        self.switchId = self.tabs.connect('switch-page', self.on_switch_page)
        self.showMenubarId = self.config.connect('change::showMenubar', \
            self.on_menubar_change)
            
        self.config.connect('change::avatarsInTaskbar', self.on_avatars_change)
        self.controller.msn.connect('display-picture-changed', self.on_displaypicture_change)
        
        self.connect('key-press-event', self.on_key_press)
        
        self.add(self.vbox)
        
        self.conversation.ui.hideUserList()
    
    def on_key_press(self, widget, event):
        if (event.state & gtk.gdk.CONTROL_MASK) and \
           (event.keyval == gtk.keysyms.Tab or \
           event.keyval == gtk.keysyms.ISO_Left_Tab):
            if event.state & gtk.gdk.SHIFT_MASK:
                self.cycleTabs(-1)
            else:
                self.cycleTabs(1)
            return True
    
    def on_focus_in_event(self, widget, event):
        self.unsetUrgency()
        self.conversation.ui.input.grabFocus()

    def on_size_alloc(self, widget, allocation):
        self.width = allocation.width
        self.height = allocation.height
        
    def on_key_close_tab(self, accelGroup, window, keyval, modifier):
        '''Catches events like Ctrl+W and closes current tab'''
        self.closeTab(self.conversation)
        
    def on_key_change_tab(self, accelGroup, window, keyval, modifier):
        '''Catches alt+number and shows tab number-1  '''
        pages = self.tabs.get_n_pages()
        new = keyval - gtk.keysyms._0 - 1   # WTF?
        if new < pages:
            self.showTab(new)
        
    def on_key_cycle_tabs(self, accelGroup, window, keyval, modifier):
        '''Catches events like Ctrl+AvPag and consequently changes current 
        tab'''

        if not modifier == gtk.gdk.CONTROL_MASK:
            return
            
        if keyval == gtk.keysyms.Page_Down:
            self.cycleTabs(1)
        elif keyval == gtk.keysyms.Page_Up:
            self.cycleTabs(-1)
    
    def on_menubar_change(self, _config, value, oldValue):
        '''handler for showMenubar config change'''
        if value == '1':
            self.menu.show_all()
        else:
            accelGroup = self.accelGroup
            accelGroup.disconnect_key(ord('L'), gtk.gdk.CONTROL_MASK)
            accelGroup.disconnect_key(ord('Q'), gtk.gdk.CONTROL_MASK)
            accelGroup.disconnect_key(ord('I'), gtk.gdk.CONTROL_MASK)
            accelGroup.disconnect_key(ord('S'), gtk.gdk.CONTROL_MASK)
            accelGroup.connect_group(ord('L'), gtk.gdk.CONTROL_MASK, \
                gtk.ACCEL_LOCKED, self.clearOutputText)
            accelGroup.connect_group(ord('Q'), gtk.gdk.CONTROL_MASK, \
                gtk.ACCEL_LOCKED, self.menu.onCloseActivate)
            accelGroup.connect_group(ord('I'), gtk.gdk.CONTROL_MASK, \
                gtk.ACCEL_LOCKED, self.menu.onInviteActivate)
            accelGroup.connect_group(ord('S'), gtk.gdk.CONTROL_MASK, \
                gtk.ACCEL_LOCKED, self.menu.onSendFileActivate)
            self.menu.hide()

    def on_avatars_change(self, _config, value, oldValue):
        '''handler for avatarsInTaskbar config change'''
        self.set_icon(self.conversation.getWindowIcon())

    def on_displaypicture_change(self, _msnp, _switchboard, _msnobj, email):
        if self.conversation.switchboard.firstUser == email:
            self.set_icon(self.conversation.getWindowIcon())

    def show_invite_dialog(self, conversation=None):
        '''show the invite dialog for conversation, if conversation is 
        None, then the current conversation is used'''
        if not self.controller or not self.controller.contacts \
            or not self.controller.msn:
            return

        conversation = conversation or self.conversation

        online_contacts = self.controller.contacts.get_online_list()
        dictionary = {}
        
        for i in online_contacts:
            dictionary[i.account] = i
        
        members = {}
        
        for mail in conversation.getMembers() + [conversation.getUser()]:
            contact = self.controller.getContact(mail)
            if contact:
                members[mail] = contact
            
        ConversationUI.InviteWindow(self.controller, None, dictionary, 
            members, self.controller.theme, conversation.inviteUser).run()

    def cycleTabs(self, cycle):
        '''Move <cycle> pages to the right, if it's negative it moves
        to the left. Usually it's 1 or -1'''
        pages = self.tabs.get_n_pages()
        active = self.tabs.get_current_page()
        self.showTab((active + cycle) % pages)
        self.menu.update()
        
    def showTab(self, num):
        '''put the num tab as the selected one
        use this only if you want to explicitly set a tab NUMBER'''
        self.on_switch_page(None, None, num)
        self.tabs.set_current_page(num)
    
    def on_switch_page(self, _notebook, _page, page_num):
        '''callback for the switch-page signal'''
        
        self.conversation.setIsCurrent(False)
        self.conversation = self.tabs.get_nth_page(page_num).parentConversation
        
        self.set_icon(self.conversation.getWindowIcon())
           
        self.update_title()   
        self.conversation.setIsCurrent(True)
        self.conversation.ui.update()
    
    def update_title(self):
        '''change window title to parsed title of current conversation'''
        self.set_title(emesenelib.common.unescape(
            self.controller.unifiedParser.getParser(
            self.conversation.getTitle()).get()))
    
    def openTab(self, conversation):
        '''open a new tab, that represent a new conversation'''
        ui = conversation.ui
        tabWidget = ui.tabWidget
                
        eventbox = gtk.EventBox()
        eventbox.set_visible_window(False)
        eventbox.add(tabWidget)
        eventbox.connect('event', self.tabs_event, ui)

        tabNum = self.tabs.append_page(child=ui, tab_label=eventbox)
        
        try:
            self.tabs.set_tab_reorderable(ui, True)
        except AttributeError:
            #print 'Your "stable" system sucks because it's filled with'
            #print 'outdated libs. Please upgrade to a decent distro. '
            # :P
            pass

        self.conversations.append(conversation)

        #add the new tab label to notebook menu popup
        self.addOrUpdateMenuPopupItem(tabNum, conversation)

        # expand and fill
        self.tabs.set_tab_label_packing(ui, True, True, gtk.PACK_START)

        self.tabs.set_show_tabs(len(self.conversations) > 1)
        
        if not self.is_active() or len(self.conversations) == 0:
            self.tabs.set_current_page(tabNum)
    
    def addOrUpdateMenuPopupItem(self, tabNum, conversation):
        '''add the new tab label to notebook popup menu'''
        
        page = self.tabs.get_nth_page(tabNum)
        hbox = gtk.HBox()
        hbox.set_spacing(6)
        image = gtk.Image()
        image.set_from_pixbuf(self.controller.theme.getImage('online'))
        hbox.pack_start(image, False, False)
        hbox.pack_start(gtk.Label(conversation.getTitle()))
        hbox.show_all()
        self.tabs.set_menu_label(page, hbox)

    def closeTab(self, conversation):
        '''Close a conversation'''
        num = self.tabs.page_num(conversation.ui)
        self.tabs.remove_page(num)

        #erase the conversation
        conversation.setStatus('closed')
        conversation.setClosed(True)
        conversation.ui.close()
        self.conversations.remove(conversation)
        
        # what do we have now?
        if len(self.conversations) == 0:
            # nothing
            self.destroy()
        else:
            # another tab, set is as current
            self.on_switch_page(None, None, self.tabs.get_current_page())
            self.tabs.set_show_tabs(len(self.conversations) > 1)
    
    def tabs_event(self, _widget, event, ui):
        if event.type == gtk.gdk.BUTTON_PRESS and event.button == 2:
            self.closeTab(ui.parentConversation)
    
    def clearOutputText(self, *args):
        self.conversation.textBuffer.set_text('')

    def show(self):
        gtk.Window.show(self)
        self.notOpen = False

    def close(self, *args):
        '''close the window and set the attribute self.closed to true'''

        self.hide()
        
        for conv in self.conversations:
            conv.setClosed(True)

        self.closed = True
        
        self.controller.config.user['convWindowWidth'] = self.width
        self.controller.config.user['convWindowHeight'] = self.height

        self.config.disconnect(self.showMenubarId)
        self.tabs.disconnect(self.switchId)

        return True

    def isClosed(self):
        '''return True if the window has been hided'''
        return self.closed

    def scrollToBottom(self, tab_num = None):
        self.conversation.ui.scrollToBottom()

    def addChildAtAnchor(self, child, anchor, tab_num = None):
        '''add a child to an anchor in the current output if tab_num is None, otherwise add it in the tab'''
        # TODO: this method should be in conversation(ui)
        if tab_num is None:
            textview = self.conversation.ui.textview
        else:
            textview = self.tabs.get_nth_page(tab_num).textview
        textview.add_child_at_anchor(child, anchor)

    def setUrgency(self):
        '''try to set the urgency hint to the window or blink the tray icon'''

        if not self.has_toplevel_focus():
            try:
                if self.controller.config.user['blinkTrayIcon']:
                    self.controller.trayIcon.tray.set_blinking(True)
            except:
                pass
            try:
                if self.controller.config.user['statusbarHighLight']:
                    self.set_urgency_hint(True) 
            except:
                pass

    def unsetUrgency(self):
        '''try to unset the urgency hint to the window and
        turn off tray icon blink'''

        try:
            self.set_urgency_hint(False)
            self.controller.trayIcon.tray.set_blinking(False) 
            self.conversation.ui.update()
        except:
            # errors related to getting disconnected don't really matter
            pass
        
    def changeFont(self):
        '''opens the font selection dialog'''
        fontDesc = pango.FontDescription(self.config.user['fontFace'])
        if self.config.user['fontItalic']:
            uStyle = pango.STYLE_ITALIC
        else:
            uStyle = pango.STYLE_NORMAL
        
        if self.config.user['fontBold']:
            uWeight = pango.WEIGHT_BOLD
        else:
            uWeight = pango.WEIGHT_NORMAL
        
        fontDesc.set_style(uStyle)
        fontDesc.set_weight(uWeight)
        fontDesc.set_size(self.config.user['fontSize'] * pango.SCALE)

        fontDialog = gtk.FontSelectionDialog(_('Choose a font'))
        fontDialog.set_font_name(fontDesc.to_string())
        response = fontDialog.run()

        if response == gtk.RESPONSE_OK:
            pangoDesc = pango.FontDescription(fontDialog.get_font_name())
            font = pangoDesc.get_family()
            style = pangoDesc.get_style()
            weight = pangoDesc.get_weight()

            italic = bold = False

            if style == pango.STYLE_ITALIC:
                italic = True
            if weight == pango.WEIGHT_BOLD:
                bold = True

            self.controller.emit('font-changed', font, bold, italic, pangoDesc.get_size() / pango.SCALE)
            
        fontDialog.destroy()

    def changeColor(self):
        '''opens the color selection dialog'''
        colorDialog = gtk.ColorSelectionDialog(_('Choose a color'))
        colorDialog.colorsel.set_has_palette(True)
        colorDialog.colorsel.set_current_color(\
            gtk.gdk.color_parse(self.config.user['fontColor']))
        response = colorDialog.run()

        if response == gtk.RESPONSE_OK:
            color = colorDialog.colorsel.get_current_color()
            
            red = color.red >> 8
            green = color.green >> 8
            blue = color.blue >> 8
            
            self.controller.emit('color-changed', \
                '#%02X%02X%02X' % (red, green, blue))
            
        colorDialog.destroy()

    def send_file_dialog(self):
        '''Displays a dialog to choose a file to send'''
        if not self.controller or not self.controller.contacts \
            or not self.controller.msn:
            return

        dialog = gtk.FileChooserDialog(_('Send file'),
            buttons=(gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL,
                     gtk.STOCK_OPEN, gtk.RESPONSE_OK))
        dialog.set_current_folder(os.path.expanduser('~'))

        response = dialog.run()

        if response == gtk.RESPONSE_OK:
            self.conversation.sendFile(dialog.get_filename())
        dialog.destroy()
            

class ConversationWindowMenu(gtk.MenuBar):
    '''This class represent the menu in the conversation window, i define
    it here because no other module here cares about this one.'''

    def __init__(self, parentConversationWindow, controller):
        '''Contructor'''

        gtk.MenuBar.__init__(self)

        self.parentConversationWindow = parentConversationWindow
        self.controller = controller
        accelGroup = parentConversationWindow.accelGroup
        
        # --- 'Conversation' menu
        
        conversationMenu = gtk.Menu()
        conversationMenuItem = gtk.MenuItem(_("_Conversation"))
        conversationMenuItem.set_submenu(conversationMenu)

        self.inviteMenuItem = gtk.ImageMenuItem(_('_Invite'))
        self.inviteMenuItem.set_image(gtk.image_new_from_stock(gtk.STOCK_ADD,
                                    gtk.ICON_SIZE_MENU))
        self.inviteMenuItem.connect('activate', self.onInviteActivate)
        self.inviteMenuItem.add_accelerator('activate', accelGroup, ord('I'), 
            gtk.gdk.CONTROL_MASK, gtk.ACCEL_VISIBLE)

        self.sendFileMenuItem = gtk.ImageMenuItem(_("_Send file"))
        self.sendFileMenuItem.set_image(gtk.image_new_from_stock(gtk.STOCK_GOTO_TOP,
                                    gtk.ICON_SIZE_MENU))
        self.sendFileMenuItem.connect('activate', self.onSendFileActivate)
        self.sendFileMenuItem.add_accelerator('activate', accelGroup, ord('S'), 
            gtk.gdk.CONTROL_MASK, gtk.ACCEL_VISIBLE)
        
        clearMenuItem = gtk.ImageMenuItem(_('C_lear Conversation'))
        clearMenuItem.set_image(gtk.image_new_from_stock(gtk.STOCK_CLEAR,
                                    gtk.ICON_SIZE_MENU))
        clearMenuItem.connect('activate',
                                parentConversationWindow.clearOutputText)
        clearMenuItem.add_accelerator('activate', accelGroup, ord('L'), 
            gtk.gdk.CONTROL_MASK, gtk.ACCEL_VISIBLE)
        
        closeMenuItem = gtk.ImageMenuItem(_('Close all'))
        closeMenuItem.set_image(gtk.image_new_from_stock(gtk.STOCK_CLOSE,
                                gtk.ICON_SIZE_MENU))
        closeMenuItem.connect('activate', self.onCloseActivate)
        closeMenuItem.add_accelerator('activate', accelGroup, ord('Q'), 
            gtk.gdk.CONTROL_MASK, gtk.ACCEL_VISIBLE)
        
        conversationMenu.add(self.inviteMenuItem)
        conversationMenu.add(self.sendFileMenuItem)
        conversationMenu.add(clearMenuItem)
        conversationMenu.add(gtk.SeparatorMenuItem())
        conversationMenu.add(closeMenuItem)
        self.add(conversationMenuItem)       
        
        # --- Format menu
        formatMenu = gtk.Menu()
        formatMenuItem = gtk.MenuItem(_("For_mat"))
        formatMenuItem.set_submenu(formatMenu)
        
        fontMenuItem = gtk.ImageMenuItem(gtk.STOCK_SELECT_FONT)
        colorMenuItem = gtk.ImageMenuItem(gtk.STOCK_SELECT_COLOR)

        fontMenuItem.connect('activate', self.onFontActivate)
        colorMenuItem.connect('activate', self.onColorActivate)
        
        formatMenu.add(fontMenuItem)
        formatMenu.add(colorMenuItem)
        self.add(formatMenuItem)

    def onCloseActivate(self, *args):
        '''This method is called when Close is activated on the menu'''
        self.parentConversationWindow.close()
    
    def onInviteActivate(self, *args):
        '''This method is called when Invite is activated on the menu'''
        self.parentConversationWindow.show_invite_dialog()

    def onSendFileActivate(self, *args):
        '''This method is called when Invite is activated on the menu'''
        self.parentConversationWindow.send_file_dialog()

    def onLogActivate(self, check):
        '''This method is called when Log is activated on the menu'''
        self.parentConversationWindow.conversation.setDoLog(
            check.get_active())

    def onFontActivate(self, *args):
        '''This method is called when font is activated on the menu'''
        self.parentConversationWindow.changeFont()

    def onColorActivate(self, *args):
        '''This method is called when color is activated on the menu'''
        self.parentConversationWindow.changeColor()

    def update(self, *args):
        '''This method is called when a switchboard error is detected'''
        self.inviteMenuItem.set_sensitive(False)
        self.sendFileMenuItem.set_sensitive(False)



# -*- coding: utf-8 -*-
#
#   This file is part of emesene.
#
#    Emesene is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    emesene is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with emesene; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

IFACE_NAME = 'net.sacredchao.QuodLibet'
IFACE_PATH = '/net/sacredchao/QuodLibet'

# interesting methods:
#  Next() - Previous() - Pause() - Play() - PlayPause()

import os
import CurrentSong

class QuodLibet( CurrentSong.DbusBase ):

    def __init__( self ):
        CurrentSong.DbusBase.__init__( self, IFACE_NAME, self.setInterface )
        try: self.iface
        except: self.iface = None
        
    def setInterface( self ):
        self.iface = self.bus.get_object( IFACE_NAME, IFACE_PATH )
        
    def isPlaying( self ):
        return self.isNameActive( IFACE_NAME )
        
    def check( self ):
        if not self.iface:
            return False
    
        artist = ''
        title = ''
        album = ''
        filename = ''
        if self.isPlaying():
            data = dict(self.iface.CurrentSong())
            if 'title' in data: title = data['title']
            if 'album' in data: album = data['album']
            if 'artist' in data: artist = data['artist']
            if '~filename' in data:
                filename = data['~filename']
                filename = filename[filename.rfind('/') +1:-1]
        
        if self.artist != artist or \
           self.title != title or \
           self.filename != filename:
            self.artist = artist
            self.title = title
            self.album = album
            self.filename = filename
            return True

        return False
        
    def getStatus( self ):
        '''check if everything is OK to start the plugin
        return a tuple whith a boolean and a message
        if OK -> ( True , 'some message' )
        else -> ( False , 'error message' )'''
        
        if os.name != 'posix':
            return (False, _('This plugin only works in posix systems'))
   
        if not self.iface:
            return (False, _('You don\'t have dbus. '
                'Try to install dbus and/or use a dbus enabled '
                'version of quodlibet.'))
        
        return ( True, 'Ok' )



# -*- coding: utf-8 -*-

#   This file is part of emesene.
#
#    Emesene is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    emesene is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with emesene; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

'''This module provides "interfaces to connect to external hosts",
including only MSNP sockets by now. (future: http[s], DC)
'''

import sys
import time
import Queue
import socket
import select
import gobject
import httplib
import threading
import traceback

# faster than concatenation
try:
    from cStringIO import StringIO
except ImportError:
    from StringIO import StringIO

import common

# http proxy works better with this info :D
PAYLOAD_COMMANDS = ['UBX', 'GCF', 'MSG', 'NOT']
IGNORE_COMMANDS = ['SBS']

class BaseConnection(gobject.GObject):
    '''This is the base class to all sockets (including http method ones),
    and higher-level connections (http or https)
    The signals below behave similar to io_add_watch in a normal socket, but
    exposing a simple, unified interface for all different socket types'''
    # Prototype weak gobject implementation included.
    
    __gsignals__ = {
        'input': (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, ()),
        'hangup': (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, ()),
        'output': (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, ()),
    }
    
    def __init__(self):
        gobject.GObject.__init__(self)
        self._hung = False
        self._signals = {'input': [], 'hangup': [], 'output': []}

    def _update_events(self):
        pass
    
    def connect(self, name, *args):
        '''Connects a signal and stores the identifier'''
        retval = gobject.GObject.connect(self, name, *args)
        self._signals[name].append(retval)
        self._update_events()
        return retval

    def disconnect(self, identifier, update=True):
        if identifier:
            gobject.GObject.disconnect(self, identifier)
        for signal in self._signals:
            if identifier in self._signals[signal]:
                self._signals[signal].remove(identifier)
        if update:
            self._update_events()

    def disconnectAll(self):
        '''Disconnects all the signal ids stored'''
        for identifiers in self._signals.values():
            for identifier in identifiers:
                self.disconnect(identifier, update=False)
        self._update_events()
    
    def hangup(self):
        if not self._hung:
            self._hung = True
            self.emit('hangup')
            self.close()
            self.disconnectAll()

class BaseSocket(BaseConnection):
    '''This is the base class to Socket and HTTPSocket
    Some high-level methods are implemented here too'''
    
    # override me
    ping_enabled = None

    def __init__(self, host, port):
        BaseConnection.__init__(self)

        # the buffer to store the received data
        self._buffer = ''
        self._host = host
        self._port = port
        
        # transaction ID
        self.tid = 1

        # ping times
        self.lastsent = int(time.time())
        self.lastpng = 0
        self.lastqng = 0

        # max time in seconds between PNG and QNG
        self.max_ping_delay = 10
        self.secs_between_pings = 20

        # on login until someone complains
        self.on_login = True
    
    def sendCommand(self, command, params=''):
        '''send a command through the socket and increment the tid'''
        
        if params != "":
            text = "%s %s %s\r\n" % (command, self.tid, params)
        else:
            text = "%s %s\r\n" % (command, self.tid)
            
        self.tid += 1
        self.send(text)
        return self.tid
        
    def sendPayloadCommand(self, command, params, payload):
        '''send a command that has an indicator of the body length'''
            
        if type(payload) == unicode:
            payload = str(payload.encode('utf-8'))
        
        text = command + " " + str(self.tid) + " "
        if params != '':
            text += params + " "
        text += str(len(payload)) + '\r\n' + payload
        
        self.tid += 1
        self.send(text)
        return self.tid
        
    def send(self, text):
        pass

    def receiveCommand(self):
        return ('', 0, '')
        
    def receivePayload(self):
        return ''

    def close(self):
        pass

    def ping(self):
        pass

    def onQng(self, tid):
        pass

    def select(self):
        return ([], [], [True])   # error!

    def establish(self, timeout=None):
        pass

class Socket(BaseSocket):
    '''This represents a MSNP protocol socket over plain TCP'''
    
    ping_enabled = True

    def __init__(self, host, port):
        BaseSocket.__init__(self, host, port)
        
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.socket.setblocking(0)
        self._event_id = 0
        self._watch_mask = 0
        self._update_events()
        
    def _update_events(self):
        '''Calls io_add_watch again if the mask changed
        (by connecting/disconnecting signals)'''
        mask = 0

        # check who's connected and enable masks
        if len(self._signals['input']):
            mask |= gobject.IO_IN
        if len(self._signals['output']):
            mask |= gobject.IO_OUT
        if len(self._signals['hangup']):
            mask |= gobject.IO_HUP | gobject.IO_ERR

        if mask == self._watch_mask:
            return
        self._watch_mask = mask

        if self._event_id != 0:
            gobject.source_remove(self._event_id)
        self._event_id = 0

        if mask:
            self._event_id = gobject.io_add_watch(self.socket,
                mask, self._event)
    
    def _event(self, fd, condition):
        if self._isOnError() or (condition & (gobject.IO_ERR | gobject.IO_HUP)):
            self.hangup()
        elif (condition & gobject.IO_IN):
            self.emit('input')
        elif (condition & gobject.IO_OUT):
            self.emit('output')
        return True
    
    def send(self, text):
        '''Sends raw bytes'''
        common.debug(">>> " + text, 'socket')

        self.lastcmd = int(time.time())
        try:
            # block for 60 seconds waiting for output ready
            if not len(select.select([], [self.socket], [], 60)[1]):
                raise socket.error
            retval = self.socket.send(text)
        except Exception, e:
            common.debug("Error on send" + str(e), 'socket')
            self.hangup()
            return ''

        return retval

    def _recv(self, bytes):
        '''Receives raw bytes'''
        if not self._canReceive():
            return ''
        
        try:
            # block for 60 seconds waiting for data to read
            # note the subtle differences with the select above
            if not len(select.select([self.socket], [], [], 60)[0]):
                raise socket.error
            return self.socket.recv(bytes)
        except Exception, e:
            common.debug("Error on recv" + str(e), 'socket')
            self.hangup()
            return None
    
    def _receiveLine(self):
        '''Receives data from the socket and return it'''
        
        buffer = StringIO()
        
        chunk = self._recv(1)
        if chunk is None: return ''
        while chunk != '\n' and chunk != '':
            buffer.write(chunk)
            chunk = self._recv(1)
            if chunk is None: return ''

        message = buffer.getvalue().rstrip('\r')
        common.debug("<<< " + message, 'socket')
        self.lastcmd = int(time.time())
        return message
        
    def receiveCommand(self):
        '''Parses the receive() return value
        in a (command, tid, params) tuple'''
        list = self._receiveLine().split(' ', 2)
        
        command, tid, params = ('', '0', '')
        
        if len(list) >= 1: command = list[0]
        if len(list) >= 2: tid = list[1]
        if len(list) >= 3: params = list[2]
        
        return (command, tid, params)
        
    def receivePayload(self, length):
        '''receive a number of characters from an earlier command'''
        
        buffer = []
        received = 0
        
        while received < length:
            chunk = self._recv(length - received)
            if chunk is None: return ''
            received += len(chunk)  # not concatenation
            buffer.append(chunk)
        
        return ''.join(buffer)

    def ping(self):
        '''ping the server if idle, errors are handled by core.py'''
        if not self.on_login and \
           (int(time.time()) - self.lastcmd) > self.secs_between_pings:
            self.send('PNG\r\n')
            self.lastpng = int(time.time())

    def onQng(self, tid):
        '''called by core.process()'''
        self.lastqng = int(time.time())
        self.max_ping_delay = int(int(tid) * 85 / 100)
        self.secs_between_pings = int(int(tid) * 9 / 10)
        
    def establish(self, timeout=None):
        '''establish is another name for connect :P
        connect to the host an port given in the contructor
        throw an exception if cannot connect'''
        self.socket.settimeout(timeout)
        try:
            self.socket.connect((self._host, self._port))
        except:
            self.socket.settimeout(None)
            raise
        self.socket.settimeout(None)
    
    def close(self):
        self.socket.close()
        self.disconnectAll()

    def _canSend(self):
        '''Calls select() to determine if the socket is ready to write'''
        fds = select.select([self.socket], [], [], 0)
        return len(fds[0]) == 1
    
    def _canReceive(self):
        '''Calls select() to determine if the socket is ready to read''' 
        fds = select.select([], [self.socket], [], 0)
        return len(fds[1]) == 1

    def _isOnError(self):
        '''Determines if the socket is on error'''

        if self._hung:
            return True

        try:
            fds = select.select([], [], [self.socket], 0)
            return len(fds[2]) == 1
        except:
            return True

    def select(self):
        '''return a select.select like tuple'''
        return select.select([self.socket], [self.socket], [], 0)


# hm... i smell cofee here..
# ..is it java?
class Proxy:
    '''Just a class to hold the proxy data'''
    
    def __init__(self, host='', port='80', \
        user='', password=''):
        
        '''constructor of the class, just set the class attributes'''
        self.host = host
        self.port = str(port)
        self.user = user
        self.password = password

class HTTPSocket(BaseSocket):
    '''This class represent an abstraction of HTTPMethod that has the
    same methods that the Socket class to make easy the use of this
    on core class, it behaves the same way'''
    
    ping_enabled = False

    def __init__(self, host, port, proxy=None, serverType='NS'):
        BaseSocket.__init__(self, host, port)
        
        self.inQueue = Queue.Queue(0)
        self.outQueue = Queue.Queue(0)
        
        self.responseBuffer = ''
        self.httpRequestManager = HttpRequestManager(self,
            host, port, proxy, serverType)
        self.httpRequestManager.start()
        
        self.hasReceived = False
        
    def send(self, text):
        '''send text with no format, this is usefull for commands like OUT\r\n
        also its used internally'''
        
        common.debug(">>> " + text, 'socket')
        self.inQueue.put(text)
    
    def _getBuffer(self, wait=False):
        if self.responseBuffer == '':
            try:
                newbuffer = self.outQueue.get(True, 0.01)
            except Queue.Empty:
                return ''
            self.responseBuffer = newbuffer
        
    def receiveCommand(self):
        '''Parses the receive() return value
        in a (command, tid, params) tuple'''
        
        self._getBuffer()
        if self.responseBuffer.find('\r\n') != -1:
            received, self.responseBuffer = \
                self.responseBuffer.split('\r\n', 1)
        else:
            received = self.responseBuffer
            self.responseBuffer = ''

        common.debug("<<< " + received, 'socket')
        self.lastcmd = int(time.time())
        
        list = received.split(' ', 2)

        command, tid, params = ('', '0', '')
        if len(list) >= 1: command = list[0]
        if len(list) >= 2: tid = list[1]
        if len(list) >= 3: params = list[2]
        
        return (command, tid, params)
        
    def receivePayload(self, length):
        '''receive a number of characters from an earlier command'''
        length = int(length) 
        received = 0
        buffer = []
        
        while received < length:
            self._getBuffer(True)
            bytes = length - received
            chunk, self.responseBuffer = self.responseBuffer[:bytes], \
                                         self.responseBuffer[bytes:]
            received += len(chunk) + 4  # not concatenation
            buffer.append(chunk)
        
        return ''.join(buffer)

    def close(self, *args):
        '''called on logout'''
        self.httpRequestManager.quit()
        self.disconnectAll()
        
    def select(self):
        if self.httpRequestManager.open:
            return ([True], [True], [])
        else:
            common.debug("select: session is closed", 'socket')
            return ([], [], [])

    def _input_handler(self):
        '''Called by a normal priority idle_add'''
        self.emit('input')
        return False

    def _hangup_handler(self):
        '''Same as above but for hangup'''
        self.emit('hangup')
        return False
        
class HttpRequestManager(threading.Thread):
    '''This class make the request via Http method in a thread'''
    
    def __init__(self, parent, destip, port, proxy, desttype):
        '''Contructor'''
        threading.Thread.__init__(self)

        self.parent = parent
        self.inQueue = parent.inQueue
        self.outQueue = parent.outQueue
        
        self.host = 'gateway.messenger.hotmail.com'
        self.port = port
        self.proxy = proxy
        self.type = desttype
        self.path = '/gateway/gateway.dll?Action=open&Server=' + desttype + \
            '&IP=' + destip
        
        self.open = True
        self.connection = None
        self.data = ''
        self.sessionID = ''

        self.dataWaiting = False # indicate if some data is waiting
                                 # if false and receive is called, we call poll
        
        if self.proxy.host:
            addr = self.proxy.host + ':' + self.proxy.port
        elif desttype == 'NS':
            addr = 'gateway.messenger.hotmail.com:80'
        else:
            addr = destip + ':80'
        self.connection = httplib.HTTPConnection(addr)
    
    def quit(self):
        '''force the thread to finish'''
        while self.inGetNoWait() != None:
            pass
        self.inQueue.put("quit")
        self.open = False
        gobject.idle_add(self.parent._hangup_handler,
            priority=gobject.PRIORITY_DEFAULT)
        common.debug(self.getName() + ' closed', 'psocket')

    def run(self):
        '''the thread main loop'''

        nextreq = ''
        while self.open:
            if nextreq:
                req = nextreq
                nextreq = ''
            else:
                # if everything is empty it may wait up to 10 seconds
                req = self.inwait(3, 'poll')
            
            common.debug(self.getName() + ' doing ' + \
                str(req)[:40].strip() + '...', 'psocket')

            if req == 'poll':
                if not self.outQueue.empty():
                    gobject.idle_add(self.parent._input_handler,
                        priority=gobject.PRIORITY_DEFAULT)
                while True:
                    req = self.inGetNoWait()
                    if req == None:
                        break
                    elif req != 'poll':
                        nextreq = req
                        break
                data = self.poll()
            elif req == 'quit':
                return
            else:
                data = self.request(payload=req)

            while len(data) > 0:
                tmp = data.split('\r\n', 1)
                if len(tmp) == 1:
                    tmp = [tmp[0], '']
                line, data = tmp
                linesplit = line.split()
                if linesplit[0] not in IGNORE_COMMANDS:
                    self.outQueue.put(line)
                if linesplit[0] in PAYLOAD_COMMANDS:
                    size = int(linesplit[-1])
                    payload, data = data[:size], data[size:]
                    self.outQueue.put(payload)

                gobject.idle_add(self.parent._input_handler,
                    priority=gobject.PRIORITY_DEFAULT)
            
        return 
    
    def poll(self):
        return self.request(path='/gateway/gateway.dll?'
            'Action=poll&SessionID=' + self.sessionID)
    
    def request(self, method='POST', path=None, payload=''):

        if path is None:
            path = self.path

        url = 'http://' + self.host + path

        def dorequest():
            self.connection.putrequest(method, url,
                                       skip_accept_encoding=True,
                                       skip_host=True)
            
            self.connection.putheader("Accept", "*/*")
            self.connection.putheader("Accept-Language", "en-us")
            self.connection.putheader("User-Agent", "MSMSGS")
            self.connection.putheader("Host", self.host)
            self.connection.putheader("Proxy-Connection", "Keep-Alive")
            self.connection.putheader("Connection", "Keep-Alive")
            self.connection.putheader("Pragma", "no-cache")
            self.connection.putheader("Content-Type",
                "application/x-msn-messenger")
            self.connection.putheader("Content-Length", str(len(payload)))
            self.connection.endheaders()
            self.connection.send(payload)
            return self.getHttpResponse()
        
        for i in range(5):
            # try up to 5 times. if it keeps failing, close
            try:
                return dorequest()
            except Exception, e:
                traceback.print_exception(*sys.exc_info())
            self.connection.close()
            time.sleep(2)
            self.connection.connect()

        self.quit()
        return 'OUT'
        
    def getHttpResponse(self):
        response = self.connection.getresponse()
        
        text = response.read().strip()
        if response.status == 500:
            common.debug("500 internal server error", "psocket")
            self.quit()
            return 'OUT'
        elif response.status != 200:
            raise httplib.HTTPException("Server not available")
        try:
            # this header contains important information
            # such as the IP of the next server we should connect
            # and the session id assigned
            data = response.getheader('x-msn-messenger', '')
            if data.count("Session=close"):
                common.debug("Session closed", "socket")
                self.quit()
                return 'OUT'

            # parse the field
            self.sessionID = data.split("; GW-IP=")[0].replace("SessionID=", "")
            self.gatewayIP = data.split("; ")[1].replace("GW-IP=", "")
            self.host = self.gatewayIP
            self.path = "/gateway/gateway.dll?SessionID=" + self.sessionID
        except Exception,e:
            common.debug('In getHttpResponse: ' + str(e), 'socket')
            common.debug('Data: "%s"' % data, 'socket')
        return text
        
    def inGetNoWait(self):
        try:
            return self.inQueue.get(True, 0.01)
        except Queue.Empty:
            return None

    def inwait(self, secs, default):
        '''waits n seconds or returns if there's something in the in queue'''
        try:
            return self.inQueue.get(True, secs)
        except Queue.Empty:
            return default




# -*- coding: utf-8 -*-

#   This file is part of emesene.
#
#    Emesene is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    emesene is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with emesene; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
'''a module to define a base object that can emit signals and call
methods that are connected to those signals'''

import gc
import weakref
import inspect

class Object(object):
    '''a base class to add signal support to our class'''

    def __init__(self):
        '''constructor'''
        self.signals = {}
        self.callbacks = {}

    def signal_add(self, name, param_count):
        '''add a supported signal to the object with the name 
        "name" and the number of parameters param_count'''

        if name in self.signals:
            raise Exception("Signal already defined")

        self.signals[name] = param_count
        self.callbacks[name] = []

    def signal_connect(self, name, callback, *args):
        '''connect the callback to the signal "name", add optional
        user parameter to the callback args.
        The number of parameters received by the callback are checked
        and an Exception is raised if the number of parameters does
        not match with the defined param_count of the signal.
        The signal is connected with a weakref, so if only reference
        to your object are from signal connections, it will be
        garbage collected'''
        if name not in self.signals:
            raise Exception("signal '%s' not defined" % (name,))

        # if the number of arguments does not match and the callback
        # doesn't have a *args parameter
        # its param_count + 1 because the first parameter is the object
        # that emits the signal
        # for callbacks of objects (methods) it's + 2 because the first
        # argument is self
        arg_spec = inspect.getargspec(callback)

        if not arg_spec[1]:
            args_num = len(arg_spec[0])
            if (inspect.isfunction(callback) and 
                args_num != self.signals[name] + 1 + len(args)):
                raise TypeError(
                    "invalid number of arguments (%d required %d found)" % \
                    (self.signals[name] + 1 + len(args), args_num))
                
            elif (inspect.ismethod(callback) and 
                args_num != self.signals[name] + 2 + len(args)):
                raise TypeError(
                    "invalid number of arguments (%d required %d found)" % \
                    (self.signals[name] + 2 + len(args), args_num))

        self.callbacks[name].append((callback, args))

    def signal_emit(self, name, *args):
        '''emit the signal of name "name" using args "args"
        args must match the number of arguments of the signal
        and name must exist'''
        if name not in self.signals:
            raise ValueError("signal '%s' not defined" % (name,))
       

        for (callback, user_args) in self.callbacks[name]: 
            all_args = list(args or []) + list(user_args or [])
            args_num = len(all_args)

            if self.signals[name] + len(user_args) != args_num:
                raise TypeError("incorrect number of parameters \
(%d required %d found)" % (self.signals[name], args_num))

            if inspect.ismethod(callback) and \
                    len(gc.get_referrers(callback.im_self)) == 1:
                #print 'deleting callback', name, callback
                del self.signals[name]
            else:
                callback(self, *all_args)
                
